'''
See COPYRIGHT.md for copyright information.
'''
import regex as re
from arelle import ModelDocument, UrlUtil, XbrlConst, XmlUtil
from arelle.ModelObject import ModelObject
from arelle.ValidateXbrlDTS import arcFromConceptQname, arcToConceptQname

inlineDisplayNonePattern = re.compile(r"display\s*:\s*none")

def checkDTSdocument(val, modelDocument, isFilingDocument):
    for hrefElt, _hrefedDoc, hrefId in modelDocument.hrefObjects:
        if hrefId:  #check scheme regardless of whether document loaded
            # check all xpointer schemes
            for scheme, _path in XmlUtil.xpointerSchemes(hrefId):
                if scheme == "element" and val.validateDisclosureSystem:
                    val.modelXbrl.error(("EFM.6.03.06", "GFM.1.01.03"),
                        _("Href %(elementHref)s may only have shorthand xpointers"),
                        edgarCode="cp-0306-Href-Xpointers",
                        modelObject=hrefElt,
                        elementHref=hrefElt.get("{http://www.w3.org/1999/xlink}href"))

    if not isFilingDocument:
        return  # not a filing's extension document

    if modelDocument.type == ModelDocument.Type.SCHEMA:
        if modelDocument.targetNamespace is not None and len(modelDocument.targetNamespace) > 85:
            l = len(modelDocument.targetNamespace.encode("utf-8"))
            if l > 255:
                val.modelXbrl.error("EFM.6.07.30",
                    _("Shorten this declaration URI from %(length)s to 255 bytes or fewer in UTF-8: %(value)s."),
                    edgarCode="du-0730-uri-length-limit",
                    modelObject=modelDocument.xmlRootElement, length=l, targetNamespace=modelDocument.targetNamespace, value=modelDocument.targetNamespace)

    if modelDocument.type in (ModelDocument.Type.SCHEMA, ModelDocument.Type.LINKBASE):
        isSchema = modelDocument.type == ModelDocument.Type.SCHEMA
        if isSchema:
            for elt in modelDocument.xmlRootElement.iter(tag="{http://www.w3.org/2001/XMLSchema}*"):
                if elt.namespaceURI == XbrlConst.xsd:
                    localName = elt.localName
                    if localName in {"element", "complexType", "simpleType"}:
                        name = elt.get("name")
                        if name and len(name) > 64:
                            l = len(name.encode("utf-8"))
                            if l > 200:
                                val.modelXbrl.error("EFM.6.07.29",
                                    _("Shorten this declaration name from %(length)s to 200 bytes or fewer in UTF-8: %(name)s"),
                                    edgarCode="du-0729-name-length-limit",
                                    modelObject=elt, element=localName, name=name, length=l)

    for elt in modelDocument.xmlRootElement.iter():
        if isinstance(elt, ModelObject):
            xlinkType = elt.get("{http://www.w3.org/1999/xlink}type")
            xlinkRole = elt.get("{http://www.w3.org/1999/xlink}role")
            if elt.namespaceURI == XbrlConst.link:
                # check schema roleTypes
                if elt.localName in ("roleType","arcroleType"):
                    uriAttr = {"roleType":"roleURI", "arcroleType":"arcroleURI"}[elt.localName]
                    roleURI = elt.get(uriAttr)
                    if len(roleURI) > 85:
                        l = len(roleURI.encode("utf-8"))
                        if l > 255:
                            val.modelXbrl.error("EFM.6.07.30",
                                _("Shorten this declaration URI from %(length)s to 255 bytes or fewer in UTF-8: %(value)s."),
                                edgarCode="du-0730-uri-length-limit",
                                modelObject=elt, element=elt.qname, attribute=uriAttr, length=l, roleURI=roleURI, value=roleURI)

                if elt.localName == "arcroleRef":
                    refUri = elt.get("arcroleURI")
                    hrefAttr = elt.get("{http://www.w3.org/1999/xlink}href")
                    hrefUri, hrefId = UrlUtil.splitDecodeFragment(hrefAttr)
                    if hrefUri not in val.disclosureSystem.standardTaxonomiesDict:
                        val.modelXbrl.error(("EFM.6.09.06", "GFM.1.04.06"),
                            _("An arcroleRef element refers to %(xlinkHref)s, an arcrole, %(refURI)s, that is not defined in a standard taxonomy. "
                              "Please recheck submission."),
                            edgarCode="fs-0906-Custom-Arcrole-Referenced",
                            modelObject=elt, refURI=refUri, xlinkHref=hrefUri)
            if modelDocument.type == ModelDocument.Type.INLINEXBRL and elt.namespaceURI in XbrlConst.ixbrlAll:
                if elt.localName == "footnote":
                    if val.validateGFM:
                        if elt.get("{http://www.w3.org/1999/xlink}arcrole") != XbrlConst.factFootnote:
                            # must be in a nonDisplay div
                            if not any(inlineDisplayNonePattern.search(e.get("style") or "")
                                       for e in XmlUtil.ancestors(elt, XbrlConst.xhtml, "div")):
                                val.modelXbrl.error(("EFM.N/A", "GFM:1.10.16"),
                                    _("Inline XBRL footnote %(footnoteID)s must be in non-displayable div due to arcrole %(arcrole)s"),
                                    modelObject=elt, footnoteID=elt.get("footnoteID"),
                                    arcrole=elt.get("{http://www.w3.org/1999/xlink}arcrole"))

                        if not elt.get("{http://www.w3.org/XML/1998/namespace}lang"):
                            val.modelXbrl.error(("EFM.N/A", "GFM:1.10.13"),
                                _("Inline XBRL footnote %(footnoteID)s is missing an xml:lang attribute"),
                                modelObject=elt, footnoteID=id)
            if xlinkType == "extended":
                if not xlinkRole or xlinkRole == "":
                    val.modelXbrl.error(("EFM.6.09.04", "GFM.1.04.04"),
                        "The %(element)s element requires an xlink:role such as the default, 'http://www.xbrl.org/2003/role/label'.  "
                        "Please recheck submission.",
                        edgarCode="fs-0904-Resource-Role-Missing",
                        modelObject=elt, element=elt.qname)
                if not val.extendedElementName:
                    val.extendedElementName = elt.qname
                elif val.extendedElementName != elt.qname and modelDocument.type == ModelDocument.Type.LINKBASE:
                    val.modelXbrl.error(("EFM.6.09.07", "GFM:1.04.07"),
                        _("Your filing contained extended type links with roles of different namesapces and local names, %(element)s "
                          "and %(element2)s in the same linkbase, %(refUrl)s.  Please recheck your submission and ensure that all "
                          "extended-type links in a single linkbase have the same namespace and local name."),
                        edgarCode="cp-0907-Linkbases-Distinct", refUrl=elt.modelDocument.basename,
                        modelObject=elt, element=elt.qname, element2=val.extendedElementName)
            elif xlinkType == "resource":
                if not xlinkRole:
                    val.modelXbrl.error(("EFM.6.09.04", "GFM.1.04.04"),
                        _("The %(element)s element requires an xlink:role such as the default, 'http://www.xbrl.org/2003/role/label'.  "
                          "Please recheck submission."),
                        edgarCode="fs-0904-Resource-Role-Missing",
                        modelObject=elt, element=elt.qname)
                elif not XbrlConst.isStandardRole(xlinkRole):
                    modelsRole = val.modelXbrl.roleTypes.get(xlinkRole)
                    if (modelsRole is None or len(modelsRole) == 0 or
                        modelsRole[0].modelDocument.targetNamespace not in val.disclosureSystem.standardTaxonomiesDict):
                        val.modelXbrl.error(("EFM.6.09.05", "GFM.1.04.05"),
                            _("The %(element)s element has a role, %(roleDefinition)s, that is not defined in a standard taxonomy, "
                              "resource labeled %(xlinkLabel)s.  Please recheck submission."),
                            edgarCode="fs-0905-Custom-Resource-Role-Used",
                            modelObject=elt, xlinkLabel=elt.get("{http://www.w3.org/1999/xlink}label"), role=xlinkRole, element=elt.qname,
                            roleDefinition=val.modelXbrl.roleTypeDefinition(xlinkRole))
            elif xlinkType == "arc":
                if elt.get("priority") is not None:
                    priority = elt.get("priority")
                    try:
                        if int(priority) >= 10:
                            val.modelXbrl.error(("EFM.6.09.09", "GFM.1.04.08"),
                                _("The priority attribute %(priority)s has a value of ten or greater, on arc %(arcElement)s from %(xlinkFrom)s to %(xlinkTo)s. "
                                  "Please change the priority to value less than 10 and resubmit."),
                                edgarCode="du-0909-Relationship-Priority-Not-Less-Than-Ten",
                                modelObject=elt,
                                arcElement=elt.qname,
                                xlinkFrom=elt.get("{http://www.w3.org/1999/xlink}from"),
                                xlinkTo=elt.get("{http://www.w3.org/1999/xlink}to"),
                                priority=priority)
                    except (ValueError) :
                        val.modelXbrl.error(("EFM.6.09.09", "GFM.1.04.08"),
                            _("The priority attribute %(priority)s has a value of ten or greater, on arc %(arcElement)s from %(xlinkFrom)s to %(xlinkTo)s. "
                              "Please change the priority to value less than 10 and resubmit."),
                            edgarCode="du-0909-Relationship-Priority-Not-Less-Than-Ten",
                            modelObject=elt,
                            arcElement=elt.qname,
                            xlinkFrom=elt.get("{http://www.w3.org/1999/xlink}from"),
                            xlinkTo=elt.get("{http://www.w3.org/1999/xlink}to"),
                            priority=priority)
                if elt.namespaceURI == XbrlConst.link:
                    if elt.localName == "presentationArc" and not elt.get("order"):
                        val.modelXbrl.error(("EFM.6.12.01", "GFM.1.06.01"),
                            _("The presentation relationship from %(conceptFrom)s to %(conceptTo)s does not have an order attribute.  "
                              "Please provide a value."),
                            edgarCode="rq-1201-Presentation-Order-Missing",
                            modelObject=elt,
                            xlinkFrom=elt.get("{http://www.w3.org/1999/xlink}from"),
                            xlinkTo=elt.get("{http://www.w3.org/1999/xlink}to"),
                            conceptFrom=arcFromConceptQname(elt),
                            conceptTo=arcToConceptQname(elt))
                    elif elt.localName == "calculationArc":
                        if not elt.get("order"):
                            val.modelXbrl.error(("EFM.6.14.01", "GFM.1.07.01"),
                                _("The calculation relationship from %(conceptFrom)s to %(conceptTo)s does not have an order attribute."),
                                edgarCode="du-1401-Calculation-Relationship-Order-Missing",
                                modelObject=elt,
                                xlinkFrom=elt.get("{http://www.w3.org/1999/xlink}from"),
                                xlinkTo=elt.get("{http://www.w3.org/1999/xlink}to"),
                                conceptFrom=arcFromConceptQname(elt),
                                conceptTo=arcToConceptQname(elt))
                        try:
                            weightAttr = elt.get("weight")
                            weight = float(weightAttr)
                            if not weight in (1, -1):
                                val.modelXbrl.error(("EFM.6.14.02", "GFM.1.07.02"),
                                    _("Weight value %(weight)s on the calculation relationship from %(conceptFrom)s to %(conceptTo)s "
                                      "must be equal to 1 or to -1. Please recheck submission."),
                                    edgarCode="fs-1402-Calculation-Relationship-Weight-Not-Unitary",
                                    modelObject=elt,
                                    xlinkFrom=elt.get("{http://www.w3.org/1999/xlink}from"),
                                    xlinkTo=elt.get("{http://www.w3.org/1999/xlink}to"),
                                    conceptFrom=arcFromConceptQname(elt),
                                    conceptTo=arcToConceptQname(elt),
                                    weight=weightAttr)
                        except ValueError:
                            val.modelXbrl.error(("EFM.6.14.02", "GFM.1.07.02"),
                                _("Weight value %(weight)s on the calculation relationship from %(conceptFrom)s to %(conceptTo)s "
                                  "must be equal to 1 or to -1. Please recheck submission."),
                                edgarCode="fs-1402-Calculation-Relationship-Weight-Not-Unitary",
                                modelObject=elt,
                                xlinkFrom=elt.get("{http://www.w3.org/1999/xlink}from"),
                                xlinkTo=elt.get("{http://www.w3.org/1999/xlink}to"),
                                conceptFrom=arcFromConceptQname(elt),
                                conceptTo=arcToConceptQname(elt),
                                weight=weightAttr)
                    elif elt.localName == "definitionArc":
                        if not elt.get("order"):
                            val.modelXbrl.error(("EFM.6.16.01", "GFM.1.08.01"),
                                _("The Definition relationship from %(conceptFrom)s to %(conceptTo)s does not have an order attribute."),
                                edgarCode="du-1601-Definition-Relationship-Order-Missing",
                                modelObject=elt,
                                xlinkFrom=elt.get("{http://www.w3.org/1999/xlink}from"),
                                xlinkTo=elt.get("{http://www.w3.org/1999/xlink}to"),
                                conceptFrom=arcFromConceptQname(elt),
                                conceptTo=arcToConceptQname(elt))



'''
Save formula linkbase into XBRL Formula (syntax) files.

See COPYRIGHT.md for copyright information.

Loads xbrl formula file syntax into formula linkbase.

To run from command line, loading formula linkbase and saving formula syntax files:

  python3 arelleCmdLine.py
     -f {DTS, instance, entry file, or formula linkbase file}
    --plugins formulaSaver.py
    --save-xbrl-formula {formula syntax output file.xf}

'''
from arelle.ViewUtilFormulae import rootFormulaObjects, formulaObjSortKey
from arelle.ModelFormulaObject import (ModelValueAssertion, ModelExistenceAssertion, ModelConsistencyAssertion,
                                       ModelAssertionSet,
                                       ModelFactVariable, ModelGeneralVariable, ModelFormula, ModelParameter,
                                       ModelFilter, ModelConceptName, ModelConceptPeriodType, ModelConceptBalance,
                                       ModelConceptCustomAttribute, ModelConceptDataType, ModelConceptSubstitutionGroup,
                                       ModelTestFilter, ModelGeneral, ModelGeneralMeasures, ModelInstantDuration,
                                       ModelDateTimeFilter, ModelSingleMeasure, ModelExplicitDimension, ModelTypedDimension,
                                       ModelEntitySpecificIdentifier, ModelEntityScheme, ModelEntityRegexpScheme,
                                       ModelEntityRegexpIdentifier, ModelMatchFilter, ModelRelativeFilter,
                                       ModelAncestorFilter, ModelParentFilter, ModelSiblingFilter, ModelNilFilter,
                                       ModelAspectCover, ModelConceptRelation,
                                       ModelCustomFunctionSignature, ModelCustomFunctionImplementation,
                                       ModelPeriod,
                                       ModelAndFilter, ModelOrFilter, ModelMessage, ModelAssertionSeverity)
from arelle import XbrlConst, XmlUtil
from arelle.Version import authorLabel, copyrightLabel
from arelle.formula import XPathParser
import os, datetime

class NotExportable(Exception):
    def __init__(self, message):
        self.message = message

def kebabCase(name):
    return "".join("-" + c.lower() if c.isupper() else c for c in name)

def strQuote(s):
    return '"' + s.replace('"', '""') + '"'

class GenerateXule:
    def __init__(self, cntlr, xfFile):
        self.modelXbrl = cntlr.modelManager.modelXbrl
        self.xfFile = xfFile
        self.xfLines = []
        self.xmlns = {}
        self.eltTypeCount = {}
        self.nextIdDupNbr = {}

        cntlr.showStatus(_("Initializing Formula Grammar"))
        XPathParser.initializeParser(cntlr.modelManager)
        cntlr.showStatus(None)

        for cfQnameArity in sorted(qnameArity
                                   for qnameArity in self.modelXbrl.modelCustomFunctionSignatures.keys()
                                   if isinstance(qnameArity, (tuple,list))):
            cfObject = self.modelXbrl.modelCustomFunctionSignatures[cfQnameArity]
            self.doObject(cfObject, None, "", set())

        rootObjects = rootFormulaObjects(self) # sets var sets up

        # put parameters at root regardless of whether linked to
        for qn, param in sorted(self.modelXbrl.qnameParameters.items(), key=lambda i:i[0]):
            self.doObject(param, None, "", set())

        for rootObject in sorted(rootObjects, key=formulaObjSortKey):
            self.doObject(rootObject, None, "", set())

        if self.xmlns:
            self.xfLines.insert(0, "")
            for prefix, ns in sorted(self.xmlns.items(), reverse=True):
                self.xfLines.insert(0, "namespace {} = \"{}\";".format(prefix, ns))

        self.xfLines.insert(0, "")
        self.xfLines.insert(0, "(: Generated from {} by Arelle on {} :)".format(self.modelXbrl.modelDocument.basename, XmlUtil.dateunionValue(datetime.datetime.now())))

        with open(xfFile, "w", encoding="utf-8") as fh:
            fh.write("\n".join(self.xfLines))

        self.modelXbrl.info("info", "saved xule file %(file)s", file=xfFile)

    @property
    def xf(self):
        return xfLines[-1]

    @xf.setter
    def xf(self, line):
        self.xfLines.append(line)

    def objectId(self, fObj, elementType):
        eltNbr = self.eltTypeCount[elementType] = self.eltTypeCount.get(elementType, 0) + 1
        _id = fObj.id
        if _id:
            idDupNbr = self.nextIdDupNbr.get(_id,"")
            self.nextIdDupNbr[_id] = (idDupNbr or 1) + 1
            _id = "{}{}".format(_id, idDupNbr)
        else:
            _id = "{}{}".format(elementType, eltNbr)
        return _id

    def doObject(self, fObj, fromRel, pIndent, visited):
        if fObj is None:
            return
        cIndent = pIndent + "   "
        if isinstance(fObj, ModelAssertionSet):
            self.xf = "{}assertion-set {} {{".format(pIndent,  self.objectId(fObj, "assertionSet"))
            for modelRel in self.modelXbrl.relationshipSet(XbrlConst.assertionSet).fromModelObject(fObj):
                self.doObject(modelRel.toModelObject, modelRel, cIndent, visited)
            self.xf = "{}}};".format(pIndent)
        elif isinstance(fObj, (ModelValueAssertion, ModelExistenceAssertion)):
            varSetType = "formula" if isinstance(fObj, ModelFormula) else "assertion"
            self.xf = "{}{} {} {{".format(pIndent, varSetType, self.objectId(fObj, varSetType))
            for arcrole in (XbrlConst.elementLabel,
                            XbrlConst.assertionSatisfiedMessage,
                            XbrlConst.assertionUnsatisfiedMessage,
                            XbrlConst.assertionUnsatisfiedSeverity
                            ):
                for modelRel in self.modelXbrl.relationshipSet(arcrole).fromModelObject(fObj):
                    self.doObject(modelRel.toModelObject, modelRel, cIndent, visited)
            if fObj.aspectModel == "non-dimensional":
                raise Exception("Non-dimensional aspect model is not supported.")
            if fObj.implicitFiltering == "false":
                self.xf = "{}no-implicit-filtering;".format(cIndent)
            if isinstance(fObj, ModelFormula):
                for attr in ("decimals", "precision", "value"):
                    if fObj.get(attr):
                        self.xf = "{}{} {{{}}};".format(cIndent, attr, fObj.get(attr))
                if fObj.get("source"):
                    self.xf = "{}source {};".format(cIndent, fObj.get("source"))
                for aspectsElt in XmlUtil.children(fObj, XbrlConst.formula, "aspects"):
                    self.xf = "{}aspect-rules{} {{".format(cIndent,
                                                           "source {}".format(aspectsElt.get("source")) if aspectsElt.get("source") else "")
                    for ruleElt in XmlUtil.children(aspectsElt, XbrlConst.formula, "*"):
                        self.doObject(ruleElt, None, cIndent + "   ", visited)
                    self.xf = "{}}};".format(cIndent)
            for arcrole in (XbrlConst.variableSetFilter,
                            XbrlConst.variableSet,
                            XbrlConst.variableSetPrecondition):
                for modelRel in self.modelXbrl.relationshipSet(arcrole).fromModelObject(fObj):
                    self.doObject(modelRel.toModelObject, modelRel, cIndent, visited)
            if isinstance(fObj, ModelValueAssertion):
                self.xf = "{}test {{{}}};".format(cIndent, fObj.viewExpression)
            elif isinstance(fObj, ModelExistenceAssertion):
                self.xf = "{}evaluation-count {{{}}};".format(cIndent, fObj.viewExpression or ". gt 0")
            self.xf = "{}}};".format(pIndent)
        elif isinstance(fObj, ModelFormula):
            raise Exception("xbrl formula (fact creation) is not supported.")
        elif isinstance(fObj, ModelConsistencyAssertion):
            raise Exception("xbrl consistency-assertion (fact comparison) is not supported.")
        elif isinstance(fObj, ModelFactVariable) and fromRel is not None:
            self.xf = "{}${} = {{".format(pIndent, fromRel.variableQname)
            if fromRel.variableQname.prefix:
                self.xmlns[fromRel.variableQname.prefix] = fromRel.variableQname.namespaceURI
            if fObj.nils == "true":
                self.xf = "nonils"

            if fObj.bindAsSequence == "true":
                self.xf = "{}as-sequence".format(cIndent)
            if fObj.matches == "true":
                self.xf = "{}matches".format(cIndent)
            if fObj.fallbackValue:
                self.xf = "{}fallback {{{}}}".format(cIndent, fObj.fallbackValue)
            for modelRel in self.modelXbrl.relationshipSet(XbrlConst.variableFilter).fromModelObject(fObj):
                self.doObject(modelRel.toModelObject, modelRel, cIndent, visited)
            self.xf = "{}}};".format(pIndent)
        elif isinstance(fObj, ModelGeneralVariable) and fromRel is not None:
            self.xf = "{}variable ${} {{".format(pIndent, fromRel.variableQname)
            if fromRel.variableQname.prefix:
                self.xmlns[fromRel.variableQname.prefix] = fromRel.variableQname.namespaceURI
            if fObj.bindAsSequence:
                self.xf = "{}as-sequence".format(cIndent)
            self.xf = "{}select {{{}}}".format(cIndent, fObj.select)
        elif isinstance(fObj, ModelParameter):
            if fromRel is not None:
                # parameter is referenced by a different QName on arc
                if fromRel.variableQname.prefix:
                    self.xmlns[fromRel.variableQname.prefix] = fromRel.variableQname.namespaceURI
                self.xf = "{}parameter ${} references ${};".format(pIndent, fromRel.variableQname, fObj.parameterQname)
            else: # root level parameter
                if fObj.parameterQname.prefix:
                    self.xmlns[fObj.parameterQname.prefix] = fObj.parameterQname.namespaceURI
                self.xf = "{}parameter {} {{".format(pIndent, fObj.parameterQname)
                if fObj.isRequired:
                    self.xf = "{}required".format(cIndent)
                self.xf = "{} select {{{}}}".format(cIndent, fObj.select)
                if fObj.asType:
                    self.xf = "{} as {{{}}}".format(cIndent, fObj.asType)
                    if fObj.asType.prefix:
                        self.xmlns[fObj.asType.prefix] = fObj.asType.namespaceURI
                self.xf = "{}}};".format(pIndent)
        elif isinstance(fObj, ModelFilter):
            if fromRel.isComplemented:
                self.xf = "{}not".format(pIndent)
            if not fromRel.isCovered and fromRel.localName == "variableFilterArc":
                self.xf = "{}non-covering".format(pIndent)
            if isinstance(fObj, ModelConceptName):
                if len(fObj.conceptQnames) == 1 and not fObj.qnameExpressions:
                    qn = next(iter(fObj.conceptQnames))
                    self.xmlns[qn.prefix] = qn.namespaceURI
                    self.xf = "{}concept-name {};".format(pIndent, qn)
                elif len(fObj.qnameExpressions) == 1 and not fObj.conceptQnames:
                    self.xf = "{}concept-name {{{}}};".format(pIndent, fObj.qnameExpressions[0])
                else:
                    self.xf = "{}concept-name".format(pIndent)
                    for qn in fObj.conceptQnames:
                        self.xmlns[qn.prefix] = qn.namespaceURI
                        self.xf = "{}   {}".format(pIndent, qn)
                    for qnExpr in fObj.qnameExpressions:
                        self.xf = "{}   {}".format(pIndent, qnExpr)
                    self.xf = "{}   ;".format(pIndent)
            elif isinstance(fObj, ModelConceptPeriodType):
                self.xf = "{}concept-period {};".format(pIndent, fObj.periodType)
            elif isinstance(fObj, ModelConceptBalance):
                self.xf = "{}concept-balance {};".format(pIndent, fObj.balance)
            elif isinstance(fObj, (ModelConceptDataType, ModelConceptSubstitutionGroup)):
                self.xf = "{}{} {} {};".format(pIndent, kebabCase(fObj.localName),
                                               "strict" if fObj.strict == "true" else "non-strict",
                                               fObj.filterQname if fObj.filterQname else "{{{}}}".format(fObj.qnameExpression))
            elif isinstance(fObj, ModelExplicitDimension):
                members = []
                for memberElt in XmlUtil.children(fObj, XbrlConst.df, "member"):
                    members.append("member")
                    member = XmlUtil.childText(memberElt, XbrlConst.df, "qname")
                    if member:
                        member = str(member) # qname, must coerce to string
                    else:
                        member = XmlUtil.childText(memberElt, XbrlConst.df, "qnameExpression")
                        if member:
                            member = "{{{}}}".format(member)
                        else:
                            member = "$" + XmlUtil.childText(memberElt, XbrlConst.df, "variable")
                    members.append(member)
                    linkrole = XmlUtil.childText(memberElt, XbrlConst.df, "linkrole")
                    if linkrole:
                        members.append("linkrole")
                        members.append("\"{}\"".format(linkrole))
                    arcrole = XmlUtil.childText(memberElt, XbrlConst.df, "arcrole")
                    if arcrole:
                        members.append("arcrole")
                        members.append("\"{}\"".format(arcrole))
                    axis = XmlUtil.childText(memberElt, XbrlConst.df, "axis")
                    if axis:
                        members.append("axis")
                        members.append(axis)
                self.xf = "{}explicit-dimension {} {};".format(pIndent,
                                                              fObj.dimQname or ("{{{}}}".format(fObj.dimQnameExpression) if fObj.dimQnameExpression else ""),
                                                              " ".join(members))
            elif isinstance(fObj, ModelTypedDimension): # this is a ModelTestFilter not same as genera/unit/period
                self.xf = "{}typed-dimension {} {};".format(pIndent,
                                                           fObj.dimQname or ("{{{}}}".format(fObj.dimQnameExpression) if fObj.dimQnameExpression else ""),
                                                           " {{{}}}".format(fObj.test) if fObj.test else "")
            elif isinstance(fObj, ModelTestFilter):
                self.xf = "{}{} {{{}}};".format(pIndent,
                                                "general" if isinstance(fObj, ModelGeneral) else
                                                "unit-general-measures" if isinstance(fObj, ModelGeneralMeasures) else
                                                "period" if isinstance(fObj, ModelPeriod) else
                                                "entity-identifier" if isinstance(fObj, ModelIdentifier) else None,
                                                fObj.test)
            elif isinstance(fObj, ModelDateTimeFilter):
                self.xf = "{}{} {{{}}}{};".format(pIndent, kebabCase(fObj.localName), fObj.date,
                                                  " {{{}}}".format(fObj.time) if fObj.time else "")
            elif isinstance(fObj, ModelInstantDuration):
                self.xf = "{}instant-duration {} {};".format(pIndent, fObj.boundary, fObj.variable)
            elif isinstance(fObj, ModelSingleMeasure):
                self.xf = "{}unit-single-measure {};".format(pIndent,
                                                             fObj.measureQname or ("{{{}}}".format(fObj.qnameExpression) if fObj.qnameExpression else ""))
            elif isinstance(fObj, ModelEntitySpecificIdentifier):
                self.xf = "{}entity scheme {{{}}} value {{{}}};".format(pIndent, fObj.scheme, fObj.value)
            elif isinstance(fObj, ModelEntityScheme):
                self.xf = "{}entity-scheme {{{}}};".format(pIndent, fObj.scheme)
            elif isinstance(fObj, ModelEntityRegexpScheme):
                self.xf = "{}entity-scheme-pattern \"{}\";".format(pIndent, fObj.pattern)
            elif isinstance(fObj, ModelEntityRegexpIdentifier):
                self.xf = "{}entity-identifier-pattern \"{}\";".format(pIndent, fObj.pattern)
            elif isinstance(fObj, ModelMatchFilter):
                self.xf = "{}{} ${} {}{};".format(pIndent, kebabCase(fObj.localName), fObj.variable,
                                                  " dimension {}".format(fObj.dimension) if fObj.get("dimension") else "",
                                                  " match-any" if fObj.matchAny else "")
            elif isinstance(fObj, ModelRelativeFilter):
                self.xf = "{}relative ${};".format(pIndent, fObj.variable)
            elif isinstance(fObj, ModelAncestorFilter):
                self.xf = "{}ancestor {};".format(pIndent,
                                                  fObj.ancestorQname or ("{{{}}}".format(fObj.qnameExpression) if fObj.qnameExpression else ""))
            elif isinstance(fObj, ModelParentFilter):
                self.xf = "{}parent {};".format(pIndent,
                                                  fObj.parentQname or ("{{{}}}".format(fObj.qnameExpression) if fObj.qnameExpression else ""))
            elif isinstance(fObj, ModelSiblingFilter):
                self.xf = "{}sibling ${};".format(pIndent, fObj.variable)
            elif isinstance(fObj, ModelNilFilter):
                self.xf = "{}nilled;".format(pIndent)
            elif isinstance(fObj, ModelAspectCover):
                aspects = []
                for aspectElt in XmlUtil.children(fObj, XbrlConst.acf, "aspect"):
                    aspects.append(XmlUtil.text(aspectElt))
                for dimElt in XmlUtil.descendants(fObj, XbrlConst.acf, ("qname", "qnameExpression")):
                    dimAspect = qname( dimElt, XmlUtil.text(dimElt) )
                    aspects.append("exclude-dimension" if dimElt.getparent().localName == "excludeDimension" else "dimension")
                    if dimElt.localName == "qname":
                        aspects.append(str(qname( dimElt, XmlUtil.text(dimElt) )))
                    else:
                        aspects.append("{{{}}}".format(XmlUtil.text(dimElt)))
                self.xf = "{}aspect-cover {};".format(pIndent, " ".join(aspects))
            elif isinstance(fObj, ModelConceptRelation):
                conceptRelationTerms = []
                if fObj.sourceQname:
                    conceptRelationTerms.append(fObj.sourceQname)
                elif fObj.variable:
                    conceptRelationTerms.append("$" + fObj.variable)
                else:
                    conceptRelationTerms.append("{{{}}}".format(fObj.sourceQnameExpression))
                if fObj.linkrole:
                    conceptRelationTerms.append("linkrole")
                    conceptRelationTerms.append(fObj.linkrole)
                elif fObj.linkroleExpression:
                    conceptRelationTerms.append("linkrole")
                    conceptRelationTerms.append("{{{}}}".format(fObj.linkroleExpression))
                if fObj.arcrole:
                    conceptRelationTerms.append("arcrole")
                    conceptRelationTerms.append(fObj.arcrole)
                elif fObj.arcroleExpression:
                    conceptRelationTerms.append("arcrole")
                    conceptRelationTerms.append("{{{}}}".format(fObj.arcroleExpression))
                if fObj.axis:
                    conceptRelationTerms.append("axis")
                    conceptRelationTerms.append(fObj.axis)
                if fObj.generations is not None:
                    conceptRelationTerms.append("generations {}".format(fObj.generations))
                if fObj.test:
                    conceptRelationTerms.append("test")
                    conceptRelationTerms.append("{{{}}}".format(fObj.test))
                self.xf = "{}concept-relation {};".format(pIndent, " ".join(conceptRelationTerms))
            elif isinstance(fObj, (ModelAndFilter, ModelOrFilter)):
                self.xf = "{}{} {{".format(pIndent, "and" if isinstance(fObj, ModelAndFilter)else "or")
                if fObj not in visited:
                    visited.add(fObj)
                    for modelRel in self.modelXbrl.relationshipSet(XbrlConst.booleanFilter).fromModelObject(fObj):
                        self.doObject(modelRel.toModelObject, modelRel, cIndent, visited)
                    visited.remove(fObj)
                self.xf = "{}}};".format(pIndent)
        elif isinstance(fObj, ModelMessage):
            role = self.role
            self.xf = "{}{}{}{} \"{}\";".format(
                pIndent,
                "satisfied-message" if fromRel.arcrole == XbrlConst.assertionSatisfiedMessage else "unsatisfied-message",
                {XbrlConst.standardMessage : "", XbrlConst.terseMessage: "terse", XbrlConst.verboseMessage: "verbose"
                 }.get(fObj.role, ""),
                " ({})".format(fObj.xmlLang) if fObj.xmlLang else "",
                fObj.text.replace('"', '""'))
        elif isinstance(fObj, ModelAssertionSeverity):
            self.xf = "{}{} {};".format(
                pIndent,
                "unsatisfied-severity",
                fObj.level)
        elif isinstance(fObj, ModelCustomFunctionSignature):
            hasImplememntation = False
            if fObj not in visited:
                visited.add(fObj)
                for modelRel in self.modelXbrl.relationshipSet(XbrlConst.functionImplementation).fromModelObject(fObj):
                    self.doObject(modelRel.toModelObject, modelRel, pIndent, visited) # note: use pIndent as parent doesn't show
                    hasImplementation = True
                visited.remove(fObj)
            if not hasImplementation:
                self.xmlns[fObj.functionQname.prefix] = fObj.functionQname.namespaceURI
                self.xf = "{}abstract-function {}({}) as {};".format(pIndent, fObj.name,
                                                                      ", ".join(str(t) for t in fObj.inputTypes),
                                                                      fObj.outputType)
        elif isinstance(fObj, ModelCustomFunctionImplementation):
            sigObj = fromRel.fromModelObject
            self.xmlns[sigObj.functionQname.prefix] = sigObj.functionQname.namespaceURI
            self.xf = "{}function {}({}) as {} {{".format(pIndent,
                                                          sigObj.name,
                                                          ", ".join("{} as {}".format(inputName, sigObj.inputTypes[i])
                                                                    for i, inputName in enumerate(fObj.inputNames)),
                                                          sigObj.outputType)
            for name, stepExpr in fObj.stepExpressions:
                if "\n" not in stepExpr:
                    self.xf = "{}step ${} {{{}}};".format(cIndent, name, stepExpr)
                else:
                    self.xf = "{}step ${} {{".format(cIndent, name)
                    for exprLine in stepExpr.split("\n"):
                        self.xf = "{}   {}".format(cIndent, exprLine.lstrip())
                    self.xf = "{}}};".format(cIndent)
            self.xf = "{}return {{{}}};".format(cIndent, fObj.outputExpression)
            self.xf = "{}}};".format(pIndent)
        elif fObj.getparent().tag == "{http://xbrl.org/2008/formula}aspects":
            # aspect rules
            arg = ""
            if fObj.localName == "concept":
                if XmlUtil.hasChild(fObj, None, "qname"):
                    arg += " " + XmlUtil.childText(fObj, None, "qname")
                elif XmlUtil.hasChild(fObj, None, "qnameExpression"):
                    arg += " {" + XmlUtil.childText(fObj, None, "qnameExpression") + "}"
            elif fObj.localName == "entityIdentifier":
                if fObj.get("scheme"): arg += " scheme {" + fObj.get("scheme") + "}"
                if fObj.get("identifier"): arg += " identifier {" + fObj.get("identifier") + "}"
            elif fObj.localName == "period":
                if XmlUtil.hasChild(fObj, None, "forever"):
                    arg += " forever"
                if XmlUtil.hasChild(fObj, None, "instant"):
                    arg += " instant"
                    attr = XmlUtil.childAttr(fObj, None, "instant", "value")
                    if attr: arg += "{" + attr + "}"
                if XmlUtil.hasChild(fObj, None, "duration"):
                    arg += " duration"
                    attr = XmlUtil.childAttr(fObj, None, "duration", "start")
                    if attr: arg += " start {" + attr + "}"
                    attr = XmlUtil.childAttr(fObj, None, "duration", "end")
                    if attr: arg += " end {" + attr + "}"
            elif fObj.localName == "unit":
                if fObj.get("augment") == "true": arg += " augment"
            if fObj.localName in ("explicitDimension", "typedDimension"):
                arg += " dimension " + fObj.get("dimension")
            if fObj.localName in ("concept", "entityIdentifier", "period"):
                arg += ";"
            else:
                arg += " {"
            self.xf = "{}{}{}".format(pIndent, kebabCase(fObj.localName), arg)
            if fObj.localName == "unit":
                for elt in fObj.iterchildren():
                    arg = ""
                    if elt.get("source"): arg += " source " + elt.get("source")
                    if elt.get("measure"): arg += " measure {" + elt.get("measure") + "}"
                    self.xf = "{}{}{};".format(cIndent, kebabCase(elt.localName), arg)
            elif fObj.localName == "explicitDimension":
                for elt in fObj.iterchildren():
                    arg = ""
                    if XmlUtil.hasChild(elt, None, "qname"):
                        arg += " " + XmlUtil.childText(elt, None, "qname")
                    elif XmlUtil.hasChild(elt, None, "qnameExpression"):
                        arg += " {" + XmlUtil.childText(elt, None, "qnameExpression") + "}"
                    self.xf = "{}{}{};".format(cIndent, kebabCase(elt.localName), arg)
            elif fObj.localName == "typedDimension":
                for elt in fObj.iterchildren():
                    arg = ""
                    if XmlUtil.hasChild(elt, None, "xpath"):
                        arg += " xpath {" + ModelXbrl.childText(elt, None, "xpath") + "}"
                    elif XmlUtil.hasChild(elt, None, "value"):
                        arg += " value " + strQoute(XmlUtil.xmlstring(XmlUtil.child(elt, None, "value"),
                                                                      stripXmlns=True, contentsOnly=False))
                    self.xf = "{}{}{};".format(cIndent, kebabCase(elt.localName), arg)
            if fObj.localName not in ("concept", "entityIdentifier", "period"):
                self.xf = "{}}};".format(pIndent)
        # check for prefixes in AST of programs of fObj
        if hasattr(fObj, "compile") and type(fObj.compile).__name__ == "method":
            fObj.compile()
            for _prog, _ast in fObj.__dict__.items():
                if _prog.endswith("Prog") and isinstance(_ast, list):
                    XPathParser.prefixDeclarations(_ast, self.xmlns, fObj)


def saveXuleMenuEntender(cntlr, menu, *args, **kwargs):
    # Extend menu with an item for the savedts plugin
    menu.add_command(label="Save Xule  file",
                     underline=0,
                     command=lambda: saveXuleMenuCommand(cntlr) )

def saveXuleMenuCommand(cntlr):
    # save DTS menu item has been invoked
    if cntlr.modelManager is None or cntlr.modelManager.modelXbrl is None:
        cntlr.addToLog("No taxonomy loaded.")
        return

    # get file name into which to save log file while in foreground thread
    xbrlFormulaFile = cntlr.uiFileDialog("save",
            title=_("arelle - Xule file"),
            initialdir=cntlr.config.setdefault("xuleFileDir","."),
            filetypes=[(_("Xule file .xule"), "*.xule")],
            defaultextension=".xule")
    if not xbrlFormulaFile:
        return False
    import os
    cntlr.config["xbrlXuleFileDir"] = os.path.dirname(xbrlFormulaFile)
    cntlr.saveConfig()

    try:
        GenerateXule(cntlr, xbrlFormulaFile, mode)
    except Exception as ex:
        dts = cntlr.modelManager.modelXbrl
        dts.error("exception",
            _("Xule file generation exception: %(error)s"), error=ex,
            modelXbrl=dts,
            exc_info=True)

def saveXuleCommandLineOptionExtender(parser, *args, **kwargs):
    # extend command line options with a save DTS option
    parser.add_option("--save-xule-formula",
                      action="store",
                      dest="xuleFile",
                      help=_("Save Xule from formula linkbase."))

def saveXuleCommandLineXbrlRun(cntlr, options, modelXbrl, *args, **kwargs):
    # extend XBRL-loaded run processing for this option
    if getattr(options, "xuleFile", False):
        if cntlr.modelManager is None or cntlr.modelManager.modelXbrl is None:
            cntlr.addToLog("No taxonomy loaded.")
            return
        GenerateXule(cntlr, options.xuleFile, "xule")


__pluginInfo__ = {
    'name': 'Save Xule File',
    'version': '0.9',
    'description': "This plug-in adds a feature to output a Xule file from formula linkbase model objects. ",
    'license': 'Apache-2',
    'author': authorLabel,
    'copyright': copyrightLabel,
    # classes of mount points (required)
    'CntlrWinMain.Menu.Tools': saveXuleMenuEntender,
    'CntlrCmdLine.Options': saveXuleCommandLineOptionExtender,
    'CntlrCmdLine.Xbrl.Run': saveXuleCommandLineXbrlRun,
}



from mock import patch

from arelle import Cntlr


@patch('io.open')
@patch('os.path.exists')
@patch('arelle.Cntlr.Cntlr.setUiLanguage')
@patch('arelle.ModelManager.ModelManager.setLocale')
def test_cntlr_cache_enabled(_, mock_codes, mock_exists, mock_open):
    mock_codes.return_value = ['en-US', 'en_US', 'en']
    mock_exists.return_value = True
    cntlr = Cntlr.Cntlr(uiLang=None)
    assert mock_open.call_count == 3
    cntlr.saveConfig()
    assert mock_open.call_count == 4

@patch('io.open')
@patch('os.path.exists')
@patch('arelle.Cntlr.Cntlr.setUiLanguage')
@patch('arelle.ModelManager.ModelManager.setLocale')
def test_cntlr_cache_disabled(_, mock_codes, mock_exists, mock_open):
    mock_codes.return_value = ['en-US', 'en_US', 'en']
    mock_exists.return_value = True
    cntlr = Cntlr.Cntlr(uiLang=None, disable_persistent_config=True)
    assert mock_open.call_count == 1
    cntlr.saveConfig()
    assert mock_open.call_count == 1



'''
formulaGenerator generates XBRL formula linkbases for a subset of the Sphinx language.

See COPYRIGHT.md for copyright information.

Sphinx is a Rules Language for XBRL described by a Sphinx 2 Primer
(c) Copyright 2012 CoreFiling, Oxford UK.
Sphinx copyright applies to the Sphinx language, not to this software.
Workiva, Inc. conveys neither rights nor license for the Sphinx language.
'''

import time, sys, io, os.path
import regex as re
from lxml import etree
from arelle.ModelValue import QName
from .SphinxParser import (parse, astNode, astNamespaceDeclaration,
                           astFormulaRule, astReportRule, astValidationRule,
                           astQnameLiteral)
from .SphinxMethods import aggreateFunctionImplementation


def generateFormulaLB(cntlr, sphinxFiles, generatedSphinxFormulasDirectory):
    from .SphinxContext import SphinxContext
    from .SphinxValidator import validate

    msgFile = None
    sourceString = None

    # logMessage operates without an instance document or ModelXbrl, so it simulates log function

    def logMessage(severity, code, text, **kwargs):
        if "sourceFileLines" in kwargs:  # use pairs of file and line number
            fileLines = ", ".join((file + (" " + str(line)) if line else "")
                                  for file, line in kwargs["sourceFileLines"])
        elif "sourceFileLine" in kwargs:
            file, line in kwargs["sourceFileLine"]
            fileLines = file + (" " + str(line)) if line else ""
        else:
            fileLines = ""
        if not fileLines:
            fileLines = ", " + fileLines
        try:
            cntlr.addToLog("[{0}] {1}{2}".format(
                                      code,
                                      text % kwargs,
                                      fileLines))
        except KeyError as err:
            cntlr.addToLog("[{0}] {1}: Missing message parameter: {2}; {3}".format(
                                      code,
                                      text,
                                      err,
                                      fileLines))

    from arelle import XmlUtil

    sc = SphinxContext( parse(cntlr, logMessage, sphinxFiles) )
    sc.logMessage = logMessage
    validate(logMessage, sc)
    assertionIDs = set()

    for prog in sc.sphinxProgs:
        sphinxFile = prog[0].fileName
        sphinxXmlns = { "xlink": 'http://www.w3.org/1999/xlink',
                        "link": 'http://www.xbrl.org/2003/linkbase',
                        "xbrli": 'http://www.xbrl.org/2003/instance',
                        "generic": 'http://xbrl.org/2008/generic',
                        "formula": 'http://xbrl.org/2008/formula',
                        "validation": 'http://xbrl.org/2008/validation',
                        "variable": 'http://xbrl.org/2008/variable',
                        "label": 'http://xbrl.org/2008/label',
                        "ca": 'http://xbrl.org/2008/assertion/consistency',
                        "ea": 'http://xbrl.org/2008/assertion/existence',
                        "va": 'http://xbrl.org/2008/assertion/value',
                        "msg": 'http://xbrl.org/2010/message',
                        "bf": 'http://xbrl.org/2008/filter/boolean',
                        "cf": 'http://xbrl.org/2008/filter/concept',
                        "df": 'http://xbrl.org/2008/filter/dimension',
                        "gf": 'http://xbrl.org/2008/filter/general',
                        "pf": 'http://xbrl.org/2008/filter/period',
                        "uf": 'http://xbrl.org/2008/filter/unit',
                        "xfi": 'http://www.xbrl.org/2008/function/instance',
                        "xsi": 'http://www.w3.org/2001/XMLSchema-instance',
                        "xs": 'http://www.w3.org/2001/XMLSchema',
                        }
        for node in prog:
            if isinstance(node, astNamespaceDeclaration):
                sphinxXmlns[node.prefix] = node.namespace

        formulaFile = sphinxFile.rpartition(".")[0] + "-formula.xml"

        # save in generatedSphinxFormulasDirectory if specified and valid (exists)
        if generatedSphinxFormulasDirectory and os.path.isdir(generatedSphinxFormulasDirectory):
            formulaFile = os.path.join(generatedSphinxFormulasDirectory,
                                       os.path.basename(formulaFile))

        xbrlLBfile = io.StringIO('''
<nsmap>
<link:linkbase

{0}
xsi:schemaLocation="http://www.xbrl.org/2003/linkbase http://www.xbrl.org/2003/xbrl-linkbase-2003-12-31.xsd"
>
<link:arcroleRef arcroleURI='http://xbrl.org/arcrole/2008/element-label'
    xlink:href='http://www.xbrl.org/2008/generic-label.xsd#element-label'
    xlink:type='simple'/>
<link:arcroleRef arcroleURI='http://xbrl.org/arcrole/2008/variable-set'
    xlink:href='http://www.xbrl.org/2008/variable.xsd#variable-set'
    xlink:type='simple'/>
<link:arcroleRef arcroleURI='http://xbrl.org/arcrole/2008/variable-filter'
    xlink:href='http://www.xbrl.org/2008/variable.xsd#variable-filter'
    xlink:type='simple'/>
<link:arcroleRef arcroleURI='http://xbrl.org/arcrole/2008/variable-set-precondition'
    xlink:href='http://www.xbrl.org/2008/variable.xsd#variable-set-precondition'
    xlink:type='simple'/>
<link:arcroleRef arcroleURI='http://xbrl.org/arcrole/2008/consistency-assertion-formula'
    xlink:href='http://www.xbrl.org/2008/consistency-assertion.xsd#consistency-assertion-formula'
    xlink:type='simple'/>
<link:roleRef roleURI='http://www.xbrl.org/2008/role/link'
    xlink:href='http://www.xbrl.org/2008/generic-link.xsd#standard-link-role'
    xlink:type='simple'/>
<link:roleRef roleURI='http://www.xbrl.org/2008/role/label'
    xlink:href='http://www.xbrl.org/2008/generic-label.xsd#standard-label'
    xlink:type='simple'/>
<link:roleRef roleURI="http://www.xbrl.org/2010/role/message"
    xlink:type="simple"
    xlink:href="http://www.xbrl.org/2010/generic-message.xsd#standard-message"/>
<link:arcroleRef arcroleURI="http://xbrl.org/arcrole/2010/assertion-unsatisfied-message"
     xlink:type="simple"
     xlink:href="http://www.xbrl.org/2010/validation-message.xsd#assertion-unsatisfied-message"/>
<link:arcroleRef arcroleURI="http://xbrl.org/arcrole/2010/assertion-satisfied-message"
     xlink:type="simple"
     xlink:href="http://www.xbrl.org/2010/validation-message.xsd#assertion-satisfied-message"/>
<link:arcroleRef arcroleURI='http://xbrl.org/arcrole/2008/boolean-filter'
     xlink:href='http://www.xbrl.org/2008/boolean-filter.xsd#boolean-filter'
     xlink:type='simple'/>

<generic:link xlink:type="extended" xlink:role="http://www.xbrl.org/2003/role/link"/>
</link:linkbase>
</nsmap>
<!--  Generated by Arelle(r) http://arelle.org -->
'''.format('\n'.join("xmlns{0}='{1}'".format((":" + prefix) if prefix else "",
                                             namespace)
                     for prefix, namespace in sphinxXmlns.items())
           )
        )
        msgFile = os.path.basename(formulaFile)
        xmlDocument = etree.parse(xbrlLBfile,base_url=formulaFile)
        xbrlLBfile.close()
        nsmapElt = xmlDocument.getroot()
        #xmlDocument.getroot().init(self)  ## is this needed ??
        for lbElement in  xmlDocument.iter(tag="{http://www.xbrl.org/2003/linkbase}linkbase"):
            break
        for e in  xmlDocument.iter(tag="{http://xbrl.org/2008/generic}link"):
            sc.genLinkElement = e
            break

        class DocObj:  # fake ModelDocument for namespaces
            def __init__(self):
                self.xmlRootElement = lbElement
                self.xmlDocument = xmlDocument
        docObj = DocObj()

        numRules = 0
        sc.generatedVarNbr = 1
        sc.xpathCode = None
        sc.bindAsSequence = False
        sc.nodeXpathVarBindings = {}

        for node in prog:
            if isinstance(node, (astFormulaRule, astReportRule, astValidationRule)):
                # form unique ID
                sc.assertionID = node.name
                if sc.assertionID in assertionIDs:
                    for suffixNumber in range(1,10000):
                        if sc.assertionID + str(suffixNumber) not in assertionIDs:
                            sc.assertionID += str(suffixNumber)
                            break
                assertionIDs.add(sc.assertionID)
                sc.assertionElt = etree.SubElement(sc.genLinkElement,
                                                   "{http://xbrl.org/2008/assertion/value}valueAssertion",
                                                   attrib={"{http://www.w3.org/1999/xlink}type": "resource",
                                                           "{http://www.w3.org/1999/xlink}label": sc.assertionID,
                                                           "id": sc.assertionID,
                                                           "aspectModel": "dimensional",
                                                           "implicitFiltering": "true"})
                sc.assertionVarNames = {"factVar_"}
                sc.generalVarNames = {}
                if isinstance(node, astFormulaRule):
                    sc.assertionElt.set("test", xpathCode(node.expr, sc))
                    msgType = "assertion-unsatisfied-message"
                elif isinstance(node, astReportRule):
                    sc.tags["value"] = node.expr  # no test expression needed
                    msgType = "assertion-unsatisfied-message"
                elif isinstance(node, astValidationRule):
                    sc.assertionElt.set("test", "not( " + xpathCode(node.expr, sc) + " )")
                    msgType = "assertion-satisfied-message"
                genMessage(node.message, sc, msgType)

        sc.nodeXpathVarBindings.clear()

        with open(formulaFile, "w", encoding="utf-8") as fh:
            XmlUtil.writexml(fh, xmlDocument, encoding="utf-8")

        logMessage("INFO", "info:compileSphinx",
                 _("Compiled Sphinx of %(sphinxFile)s has %(numberRules)s tables in file %(formulaFile)s."),
                 sphinxFile=sphinxFiles,
                 numberRules=numRules,
                 formulaFile=formulaFile)

    logMessage = None
    sc.close()

    cntlr.showStatus("Finshed sphinx files {0}".format(", ".join(os.path.basename(f)
                                                                 for f in sphinxFiles)),
                     5000)

def evaluate(node, sc):
    if isinstance(node, astNode):
        className = node.__class__.__name__
        if className in evaluationHandler:
            return evaluationHandler[className](node, sc)
        else: # don't evaluate node
            return node
    elif isinstance(node, (tuple,list)):
        return [evaluate(item, sc) for item in node]
    elif isinstance(node, set):
        return set(evaluate(item, sc) for item in node)
    elif callable(node):
        return node()
    else:
        return node

def evaluateFunctionReference(node, sc):
    name = node.name
    if name in {"list", "set"}:
        # evaluate args
        args = [evaluate(arg, sc) for arg in node.args]
        # process args
        if name == "list":
            return list(args)
        elif name == "set":
            return set(args)
    return node

def evaluateQnameLiteral(node, sc):
    return node.value

def evaluateStringLiteral(node, sc):
    return node.text

def evaluateVariableAssignment(node, sc):
    result = evaluate(node.expr, sc)
    sc.localVariables[node.variableName] = result
    if node.tagName:
        sc.tags[node.tagName] = node.expr
    return result

def evaluateVariableReference(node, sc):
    # expand macro variables
    if node.variableName in sc.localVariables:
        return evaluate(sc.localVariables[node.variableName], sc)
    return '$' + node.variableName

def xpathCode(node, sc):
    parentXpathCode = sc.xpathCode
    sc.xpathCode = []
    genNode(node, sc)
    xpathCodeStr = " ".join(sc.xpathCode)
    sc.xpathCode = parentXpathCode
    return xpathCodeStr

def genNode(node, sc):
    if isinstance(node, astNode):
        genNodeHandler[node.__class__.__name__](node, sc)
    elif isinstance(node, (tuple,list)):
        sc.xpathCode.append('[')
        for item in node:
            genNode(item, sc)
        sc.xpathCode.append(']')
    elif isinstance(node, set):
        sc.xpathCode.append("set(")
        for item in node:
            genNode(item, sc)
        sc.xpathCode.append(")")
    elif callable(node):
        node()
    else:
        sc.xpathCode.append(str(node))

def genBinaryOperation(node, sc):
    # for now formula := generates a "eq" operation so it fails when conditoin is not met
    sc.xpathCode.append("(")
    genNode(node.leftExpr, sc)
    sc.xpathCode.append({"+": "+", "-": "-", "*": "", "/": "div",
                         "<": "lt", "<=": "le", ">": "gt", ">=": "ge",
                         "==": "eq", "!=": "ne",
                         ":=": "eq",
                         }.get(node.op,"$$$opNotImplemented$$$"))
    genNode(node.rightExpr, sc)
    sc.xpathCode.append(")")

def genFor(node, sc):
    # add a general variable for this node's expression, substitute gen var expression
    generalVarLabel = "generalVariable_" + node.name
    sc.xpathCode.append("$" + node.name)

    def genForSelectCode():
        sc.xpathCode.append("for $_{0} in {1} return ( {2} )".format(
                                node.name,
                                genNode(node.collectionExpr, sc),
                                genNode(node.expr, sc)
                                ))

    etree.SubElement(sc.genLinkElement,
                     "{http://xbrl.org/2008/variable}generalVariable",
                     attrib={"{http://www.w3.org/1999/xlink}type": "resource",
                             "{http://www.w3.org/1999/xlink}label": generalVarLabel,
                             "select": xpathCode(genForSelectCode, sc)})
    etree.SubElement(sc.genLinkElement,
                     "{http://xbrl.org/2008/variable}variableArc",
                     attrib={"{http://www.w3.org/1999/xlink}type": "arc",
                             "{http://www.w3.org/1999/xlink}arcrole": "http://xbrl.org/arcrole/2008/variable-set",
                             "{http://www.w3.org/1999/xlink}from": sc.assertionID,
                             "{http://www.w3.org/1999/xlink}to": generalVarLabel,
                             "name": node.name})

def genFunctionDeclaration(node, sc, args):
    # attempt to expand macros
    if node.functionType == "macro":
        # expand locally
        overriddenVariables = {}

        if isinstance(args, dict):
            # args may not all be used in the function declaration, just want used ones
            argDict = dict((name, value)
                           for name, value in args.items()
                           if name in node.params)
        else:  # purely positional args
            # positional parameters named according to function prototype
            if len(args) != len(node.params):
                sc.modelXbrl.log("ERROR", "sphinx.functionArgumentsMismatch",
                                 _("Function %(name)s requires %(required)s parameters but %(provided)s are provided"),
                                 sourceFileLine=node.sourceFileLine,
                                 name=node.name, required=len(node.params), provided=len(args))
                return None
            argDict = dict((paramName, args[i])
                           for i, paramName in enumerate(node.params))
        for name, value in argDict.items():
            value = evaluate(value, sc)  # resolves variable refs to same variable ref
            if name in sc.localVariables:
                overriddenVariables[name] = sc.localVariables[name]
            sc.localVariables[name] = value
        result = genNode(node.expr, sc)
        for name in argDict.keys():
            del sc.localVariables[name]
        sc.localVariables.update(overriddenVariables)
        overriddenVariables.clear()
    elif node.functionType == "function":
        # generate custom function implementation
        pass

def genFunctionReference(node, sc):
    name = node.name
    if name in sc.functions:  # user defined function
        result = genFunctionDeclaration(sc.functions[name], sc, node.args)

    isAggregateFunction = name in aggreateFunctionImplementation
    if isAggregateFunction:
        priorBindAsSequence = sc.bindAsSequence
        sc.bindAsSequence = True
    formulaFunctionSyntax = formulaFunctionSyntaxPatterns.get(name,
                              ("$notImplemented_{0}(".format(name),")"))
    if isinstance(formulaFunctionSyntax, tuple) and len(formulaFunctionSyntax) == 2:
        sc.xpathCode.append(formulaFunctionSyntax[0]) # before arguments
        for i, arg in enumerate(node.args):
            if i:
                sc.xpathCode.append(",")
            genNode(arg, sc)
        sc.xpathCode.append(formulaFunctionSyntax[1]) # after arguments
    else:
        sc.logMessage("ERROR", "sphinxFormulaGenerator:functionSyntaxPatternError",
                      _("Syntax generator error for function %(name)s."),
                      sourceFileLine=node.sourceFileLine, name=name)
    if isAggregateFunction:
        sc.bindAsSequence = priorBindAsSequence

def genHyperspaceExpression(node, sc):
    # generate fact variable and substitute it's reference in text code
    # find pri item for fact var name
    priItemName = "factVar_"
    for axis in node.axes:
        if axis.name == "primary":
            restriction = evaluate(axis.restriction, sc)
            if len(restriction) == 1:
                priItemName += cleanedName(str(restriction[0]))
            break

    if priItemName in sc.assertionVarNames:
        for suffixNumber in range(1,10000):
            if priItemName + str(suffixNumber) not in sc.assertionVarNames:
                priItemName += str(suffixNumber)
                break
    sc.assertionVarNames.add(priItemName)
    factVarLabel = sc.assertionID + "_" + priItemName
    xpathVar = "$" + factVarLabel
    sc.xpathCode.append(xpathVar)
    sc.nodeXpathVarBindings[node] = xpathVar
    etree.SubElement(sc.genLinkElement,
                     "{http://xbrl.org/2008/variable}factVariable",
                     attrib={"{http://www.w3.org/1999/xlink}type": "resource",
                             "{http://www.w3.org/1999/xlink}label": factVarLabel,
                             "bindAsSequence": ("false","true")[sc.bindAsSequence]})
    etree.SubElement(sc.genLinkElement,
                     "{http://xbrl.org/2008/variable}variableArc",
                     attrib={"{http://www.w3.org/1999/xlink}type": "arc",
                             "{http://www.w3.org/1999/xlink}arcrole": "http://xbrl.org/arcrole/2008/variable-set",
                             "{http://www.w3.org/1999/xlink}from": sc.assertionID,
                             "{http://www.w3.org/1999/xlink}to": factVarLabel,
                             "name": factVarLabel})
    # generate filters
    filterNames = {"filter_"}
    for axis in node.axes:
        filterName = "filter_"
        axisName = evaluate(axis.name, sc)
        if axisName == "primary":
            axisRestriction = evaluate(axis.restriction, sc)[0]
            filterName = "concept"
        elif isinstance(axisName, QName):
            axisRestriction = evaluate(axis.restriction, sc)[0]
            filterName = "dim_" + axisRestriction.localName
        if filterName in filterNames:
            for suffixNumber in range(1,10000):
                if filterName + str(suffixNumber) not in filterNames:
                    filterName += str(suffixNumber)
                    break
        filterNames.add(filterName)
        filterLabel = factVarLabel  + "_" + filterName
        if axis.name == "primary":
            elt = etree.SubElement(sc.genLinkElement, "{http://xbrl.org/2008/filter/concept}conceptName",
                                   attrib={"{http://www.w3.org/1999/xlink}type": "resource",
                                           "{http://www.w3.org/1999/xlink}label": filterLabel})
            etree.SubElement(sc.genLinkElement, "{http://xbrl.org/2008/variable}variableFilterArc",
                             attrib={"{http://www.w3.org/1999/xlink}type": "arc",
                                     "{http://www.w3.org/1999/xlink}arcrole": "http://xbrl.org/arcrole/2008/variable-filter",
                                     "{http://www.w3.org/1999/xlink}from": factVarLabel,
                                     "{http://www.w3.org/1999/xlink}to": filterLabel,
                                     "complement": "false",
                                     "cover": "true"})
            elt = etree.SubElement(elt, "{http://xbrl.org/2008/filter/concept}concept")
            etree.SubElement(elt, "{http://xbrl.org/2008/filter/concept}qname").text = str(axisRestriction)
        elif isinstance(axis.name, QName): # dimension
            elt = etree.SubElement(sc.genLinkElement, "{http://xbrl.org/2008/filter/dimension}explicitDimension",
                                   attrib={"{http://www.w3.org/1999/xlink}type": "resource",
                                           "{http://www.w3.org/1999/xlink}label": filterLabel})
            etree.SubElement(sc.genLinkElement, "{http://xbrl.org/2008/variable}variableFilterArc",
                             attrib={"{http://www.w3.org/1999/xlink}type": "arc",
                                     "{http://www.w3.org/1999/xlink}arcrole": "http://xbrl.org/arcrole/2008/variable-filter",
                                     "{http://www.w3.org/1999/xlink}from": factVarLabel,
                                     "{http://www.w3.org/1999/xlink}to": filterLabel,
                                     "complement": "false",
                                     "cover": "true"})
            elt = etree.SubElement(elt, "{http://xbrl.org/2008/filter/dimension}dimension")
            etree.SubElement(elt, "{http://xbrl.org/2008/variable}dimension}qname").text = str(axisName)
            elt = etree.SubElement(elt, "{http://xbrl.org/2008/filter/dimension}member")
            etree.SubElement(elt, "{http://xbrl.org/2008/variable}dimension}qname").text = str(axisRestriction)

def genIf(node, sc):
    sc.xpathCode.append("if (")
    genNode(node.condition, sc)
    sc.xpathCode.append(") then (")
    genNode(node.thenExpr, sc)
    sc.xpathCode.append(") else  (")
    genNode(node.elseExpr, sc)
    sc.xpathCode.append(")")

def genMessage(node, sc, msgType):
    def taggedVarGenVar(node):
        tagGenVarLabel = "tagGeneralVariable_" + str(sc.generatedVarNbr)
        sc.generatedVarNbr += 1
        tagGenVarName = "$" + tagGenVarLabel
        sc.nodeXpathVarBindings[node] = tagGenVarName
        etree.SubElement(sc.genLinkElement,
                         "{http://xbrl.org/2008/variable}generalVariable",
                         attrib={"{http://www.w3.org/1999/xlink}type": "resource",
                                 "{http://www.w3.org/1999/xlink}label": tagGenVarLabel,
                                 "select": xpathCode(node, sc)})
        etree.SubElement(sc.genLinkElement,
                         "{http://xbrl.org/2008/variable}variableArc",
                         attrib={"{http://www.w3.org/1999/xlink}type": "arc",
                                 "{http://www.w3.org/1999/xlink}arcrole": "http://xbrl.org/arcrole/2008/variable-set",
                                 "{http://www.w3.org/1999/xlink}from": sc.assertionID,
                                 "{http://www.w3.org/1999/xlink}to": tagGenVarLabel,
                                 "name": tagGenVarName})

    msgLabel = sc.assertionID + "_" + msgType
    # generate mesage text
    if node and node.message:
        msgstr = evaluate(node.message, sc)
    else:
        msgstr = "Result ${value}"
    text = []
    i = 0
    while True:
        j = msgstr.find("${", i)
        if j >= 0:
            text.append(msgstr[i:j]) # previous part of string
            k = msgstr.find("}", j+2)
            if k > j:
                tagstr = msgstr[j+2:k].strip()
                tag, sep, modifier = tagstr.partition(".")
                if tag == "context":
                    text.append("{{${0}@contextRef}}".format(tag))
                elif tag in sc.tags:
                    taggedNode = sc.tags[tag]
                    if taggedNode not in sc.nodeXpathVarBindings:
                        taggedVarGenVar(taggedNode) # add a gen var for this tagged expression
                    text.append("{{{0}}}".format(sc.nodeXpathVarBindings[taggedNode]))
                else:
                    text.append('"{0}" is not supported for formula message generation'.format(tagstr))
                i = k + 1
        else:
            text.append(msgstr[i:])
            break
    messageStr = ''.join(text)
    etree.SubElement(sc.genLinkElement,
                     "{http://xbrl.org/2010/message}message",
                     attrib={"{http://www.w3.org/1999/xlink}type": "resource",
                             "{http://www.w3.org/1999/xlink}label": msgLabel,
                             "{http://www.w3.org/1999/xlink}role": "http://www.xbrl.org/2010/role/message",
                             "{http://www.w3.org/XML/1998/namespace}lang": "en"
                             }).text = messageStr
    etree.SubElement(sc.genLinkElement,
                     "{http://xbrl.org/2008/generic}arc",
                     attrib={"{http://www.w3.org/1999/xlink}type": "arc",
                             "{http://www.w3.org/1999/xlink}arcrole": "http://xbrl.org/arcrole/2010/" + msgType,
                             "{http://www.w3.org/1999/xlink}from": sc.assertionID,
                             "{http://www.w3.org/1999/xlink}to": msgLabel})

def genMethodReference(node, sc):
    formulaMethodSyntax = formulaMethodSyntaxPatterns.get(node.name,
                              ("$notImplemented_{0}(".format(node.name),")"))
    if isinstance(formulaMethodSyntax, tuple) and len(formulaMethodSyntax) == 2:
        sc.xpathCode.append(formulaMethodSyntax[0]) # before arguments
        for i, arg in enumerate(node.args):
            if i:
                sc.xpathCode.append(",")
            genNode(arg, sc)
        sc.xpathCode.append(formulaMethodSyntax[1]) # after arguments
    else:
        sc.logMessage("ERROR", "sphinxFormulaGenerator:methodSyntaxPatternError",
                      _("Syntax generator error for method %(name)s."),
                      sourceFileLine=node.sourceFileLine, name=node.name)

def genNoOp(node, sc):
    pass

def genNumericLiteral(node, sc):
    sc.xpathCode.append(str(node.value))

def genQnameLiteral(node, sc):
    sc.xpathCode.append(str(node.value))

def genStringLiteral(node, sc):
    sc.xpathCode.append('"' + node.text.replace('"', '""') + '"')

def genTagAssignment(node, sc):
    genNode(node.expr, sc) # expand syntax in place
    sc.tags[node.tagName] = node.expr

def genUnaryOperation(node, sc):
    sc.xpathCode.append("(")
    sc.xpathCode.append({"+": "+",
                         "-": "-",
                         "brackets": "",
                        }.get(node.op,"$$$opNotImplemented$$$"))
    genNode(node.expr, sc)
    sc.xpathCode.append(")")

def genValuesIteration(node, sc):
    pass

def genVariableAssignment(node, sc):
    # add a general variable for this node's expression, substitute gen var expression
    generalVarLabel = "generalVariable_" + node.variableName
    sc.xpathCode.append("$" + node.variableName)
    if node.tagName:
        sc.tags[node.tagName] = node.expr

    etree.SubElement(sc.genLinkElement,
                     "{http://xbrl.org/2008/variable}generalVariable",
                     attrib={"{http://www.w3.org/1999/xlink}type": "resource",
                             "{http://www.w3.org/1999/xlink}label": generalVarLabel,
                             "select": xpathCode(node.expr, sc)})
    etree.SubElement(sc.genLinkElement,
                     "{http://xbrl.org/2008/variable}variableArc",
                     attrib={"{http://www.w3.org/1999/xlink}type": "arc",
                             "{http://www.w3.org/1999/xlink}arcrole": "http://xbrl.org/arcrole/2008/variable-set",
                             "{http://www.w3.org/1999/xlink}from": sc.assertionID,
                             "{http://www.w3.org/1999/xlink}to": generalVarLabel,
                             "name": node.variableName})

def genVariableReference(node, sc):
    # if a macro reference, substitute the actual node
    if node.variableName in sc.localVariables:
        genNode(sc.localVariables[node.variableName], sc)
    else:
        sc.xpathCode.append("$" + node.variableName)

def genWith(node, sc):
    # with-dependent code goes into a fake general variable
    withExprGeneralVarLabel = "withExpressionGeneralVariable_" + str(sc.generatedVarNbr)
    sc.generatedVarNbr += 1
    sc.xpathCode.append("$" + withExprGeneralVarLabel)

    def genWithGenVarCode():
        # partition facts according to restriction
        sc.xpathCode.append(" if (false()) then (")
        # forces dependency by fake hyperspace fact variable reference
        genNode(node.restrictionExpr)
        sc.xpathCode.append(") else (")
        for varAssignNode in node.variableAssignments:
            genNode(node.varAssignNode, sc)
        genNode(node.bodyExpr)
        sc.xpathCode.append(")")

    etree.SubElement(sc.genLinkElement,
                     "{http://xbrl.org/2008/variable}generalVariable",
                     attrib={"{http://www.w3.org/1999/xlink}type": "resource",
                             "{http://www.w3.org/1999/xlink}label": withExprGeneralVarLabel,
                             "select": xpathCode(genWithGenVarCode, sc)})
    etree.SubElement(sc.genLinkElement,
                     "{http://xbrl.org/2008/variable}variableArc",
                     attrib={"{http://www.w3.org/1999/xlink}type": "arc",
                             "{http://www.w3.org/1999/xlink}arcrole": "http://xbrl.org/arcrole/2008/variable-set",
                             "{http://www.w3.org/1999/xlink}from": sc.assertionID,
                             "{http://www.w3.org/1999/xlink}to": withExprGeneralVarLabel,
                             "name": node.variableName})

def cleanedName(name):
    return re.sub(r"\W", "_", name)

genNodeHandler = {
    "astAnnotationDeclaration":   genNoOp,
    "astBinaryOperation":         genBinaryOperation,
    "astComment":                 genNoOp,
    "astFor":                     genFor,
    "astFunctionReference":       genFunctionReference,
    "astHyperspaceExpression":    genHyperspaceExpression,
    "astIf":                      genIf,
    "astMethodReference":         genMethodReference,
    "astNamespaceDeclaration":    genNoOp,
    "astNode":                    genNoOp,
    "astNoOp":                    genNoOp,
    "astNumericLiteral":          genNumericLiteral,
    "astQnameLiteral":            genQnameLiteral,
    "astSourceFile":              genNoOp,
    "astStringLiteral":           genStringLiteral,
    "astTagAssignment":           genTagAssignment,
    "astValuesIteration":         genValuesIteration,
    "astVariableAssignment":      genVariableAssignment,
    "astVariableReference":       genVariableReference,
    "astUnaryOperation":          genUnaryOperation,
    "astWith":                    genWith,
    }

evaluationHandler = {
    "astFunctionReference":       evaluateFunctionReference,
    "astQnameLiteral":            evaluateQnameLiteral,
    "astStringLiteral":           evaluateStringLiteral,
    "astVariableAssignment":      evaluateVariableAssignment,
    "astVariableReference":       evaluateVariableReference,
    }
formulaMethodSyntaxPatterns = {
    #"abs":          _abs,
    #"add-time-period": _addTimePeriod,
    #"balance":      _balance,
    #"concept":      _concept,
    #"concepts":     _concepts,
    #"contains":     _contains,
    #"credit":       _credit,
    #"days":         _days,
    #"debit":        _debit,
    #"decimals":     _decimals,
    #"dimension":    _dimension,
    #"dts-document-locations": _dtsDocumentLocations,
    #"duration":     _duration,
    #"end-date":     _endDate,
    #"ends-with":    _endsWith,
    #"entity":       _entityMethod,
    #"exp":          _exp,
    #"has-decimals": _hasDecimals,
    #"has-precision":_hasPrecision,
    #"identifier":   _identifier,
    #"index-of":     _indexOf,
    #"instant":      _instant,
    #"is-forever":   _isForever,
    #"is-monetary":  _isMonetary,
    #"is-numeric":   _isNumeric,
    #"last-index-of":_lastIndexOf,
    #"ln":           _ln,
    #"length":       _length,
    #"local-part":   _localPart,
    #"log":          _log,
    #"log10":        _log10,
    #"lower-case":   _lowerCase,
    #"name":         _name,
    #"period":       _period,
    #"period-type":  _periodType,
    #"power":        _power,
    #"precision":    _precision,
    #"primary":      _primary,
    #"replace":      _replace,
    #"round":        _roundItem,
    #"round-by-decimals": _roundDecimals,
    #"round-by-precision": _roundPrecision,
    #"scenario":     _scenario,
    #"schema-type":  _schemaTypeMethod,
    #"scheme":       _scheme,
    #"segment":      _segment,
    #"signum":       _signum,
    #"size":         _size,
    #"sqrt":         _sqrt,
    #"start-date":   _startDate,
    #"starts-with":  _startsWith,
    #"subtract-time-period": _subtractTimePeriod,
    #"taxonomy":     _taxonomy,
    #"to-list":      _toList,
    #"to-set":       _toSet,
    #"tuple":        _tuple,
    #"unit":         _unitMethod,
    #"unknown":      _notImplemented,
    #"upper-case":   _upperCase,
    #"xbrl-type":    _xbrlTypeMethod,

    #networks
    #"concept-hypercube-all":                _conceptHypercubeAll,
    #"concept-hypercube-all-networks":       _conceptHypercubeAllNetworks,
    #"concept-hypercube-not-all":            _conceptHypercubeNotAll,
    #"concept-hypercube-not-all-networks":   _conceptHypercubeNotAllNetworks,
    #"dimension-default":                    _dimensionDefault,
    #"dimension-default-networks":           _dimensionDefaultNetworks,
    #"dimension-domain":                     _dimensionDomain,
    #"dimension-domain-networks":            _dimensionDomainNetworks,
    #"domain-member":                        _domainMember,
    #"domain-member-networks":               _domainMemberNetworks,
    #"essence-alias":                        _essenceAlias,
    #"essence-alias-networks":               _essenceAliasNetworks,
    #"general-special":                      _generalSpecial,
    #"general-special-networks":             _generalSpecialNetworks,
    #"generic-link":                         _genericLink,
    #"generic-link-networks":                _genericLinkNetworks,
    #"hypercube-dimension":                  _hypercubeDimension,
    #"hypercube-dimension-networks":         _hypercubeDimensionNetworks,
    #"network":                              _network,
    #"networks":                             _networks,
    #"parent-child":                         _parentChild,
    #"parent-child-networks":                _parentChildNetworks,
    #"requires-element":                     _requiresElement,
    #"requires-element-networks":            _requiresElementNetworks,
    #"similar-tuples":                       _similarTuples,
    #"similar-tuples-networks":              _similarTuplesNetworks,
    #"summation-item":                       _summationItem,
    #"summation-item-networks":              _summationItemNetworks,

    #network methods
    #"ancestors":                            _ancestors,
    #"ancestor-relationships":               _ancestorRelationships,
    #"arc-name":                             _arcName,
    #"arcrole":                              _arcrole,
    #"children":                             _children,
    #"concepts":                             _concepts,
    #"descendant-relationships":             _descendantRelationships,
    #"descendants":                          _descendants,
    #"extended-link-name":                   _extendedLinkName,
    #"incoming-relationships":               _incomingRelationships,
    #"outgoing-relationships":               _outgoingRelationships,
    #"parents":                              _parents,
    #"relationships":                        _relationships,
    #"role":                                 _linkRole,
    #"source-concepts":                      _sourceConcepts,
    #"target-concepts":                      _targetConcepts,

    #relationship methods
    #"order":                                _order,
    #"preferred-label":                      _preferredLabel,
    #"source":                               _source,
    #"target":                               _target,
    #"weight":                               _weight,
    }

formulaFunctionSyntaxPatterns = {
    "boolean":      ("xs:boolean(", ")"),
    #"current-date-as-utc": _todayUTC,
#    "duration":     _durationFunction,
    "entity":       ("(", ")"),
    #"forever":      _foreverFunction,
    "instant":      ("xs:date(", ")"),
    "number":       ("number(", ")"),
    #"schema-type":  _schemaTypeFunction,
    #"time-period":  _timePeriodFunction,
    #"unit":         _unitFunction,
    }

aggreativeFunctions = {
    "all":          ("every $test in (", ") satisfies $test)"),
    "any":          ("some $test in (", ") satisfies $test)"),
    "avg":          ("avg(", ")"),
    "count":        ("count(", ")"),
    "exists":       ("exists(", ")"),
    "first":        ("(", ")[1]"),
    "list":         ("(", ")"),
    "max":          ("max(", ")"),
    #"median":       _median,
    "min":          ("min(", ")"),
    "missing":      ("empty(", ")"),
    #"mode":         _mode,
    #"modes":        _modes,
    #"set":          _set,
    #"stdevp":       _stdevp,
    "sum":          ("sum(", ")"),
    #"var":          _var,
    #"varp":         _varp,
    }



'''
See COPYRIGHT.md for copyright information.
'''

import time
from math import isnan, isinf
from collections import defaultdict
from decimal import Decimal
from arelle import Locale
from arelle.PythonUtil import OrderedDefaultDict
from arelle.ValidateXbrlCalcs import ZERO, inferredDecimals, rangeValue
from arelle.Version import authorLabel, copyrightLabel
from arelle.XbrlConst import link, xbrli, xl, xlink, domainMember

calc2YYYY = "http://xbrl.org/WGWD/YYYY-MM-DD/calculation-2.0"
calc2 = {calc2YYYY}
calc2e = "http://xbrl.org/WGWD/YYYY-MM-DD/calculation-2.0/error"
sectionFact = "http://xbrl.org/arcrole/WGWD/YYYY-MM-DD/section-fact"
calc2linkroles = "{http://xbrl.org/WGWD/YYYY-MM-DD/calculation-2.0}linkroles"
summationItem = "http://xbrl.org/arcrole/WGWD/YYYY-MM-DD/summation-item" # calc2 summation-item arc role
balanceChanges = "http://xbrl.org/arcrole/WGWD/YYYY-MM-DD/balance-changes"
aggregationDomain = "http://xbrl.org/arcrole/WGWD/YYYY-MM-DD/aggregation-domain"
calc2Arcroles = (summationItem, balanceChanges, aggregationDomain)

def nominalPeriod(duration): # account for month and year lengths
    if 364 < duration.days <= 366: return 365
    if 28 <= duration.days <= 31: return 31
    return duration.days

def intervalZero():
    return (Decimal(0), Decimal(0))

NIL = "(nil)" # singleton object, use "is" to compare, not the value
NILinterval = (NIL,NIL)

def intervalValue(fact, dec=None): # value in decimals
    if fact.isNil:
        return NILinterval
    if dec is None:
        dec = inferredDecimals(fact)
    return rangeValue(fact.value, dec)

def addInterval(boundValues, key, intervalValue, weight=None):
    a, b, _inclA, _inclB = intervalValue
    if a is NIL:
        result = NILinterval
    else:
        r = boundValues[key]
        if r is NILinterval:
            return
        elif weight is not None:
            result = (r[0] + weight * a, r[1] + weight * b)
        else:
            result = (r[0] + a, r[1] + b)
    boundValues[key] = result


class ValidateXbrlCalc2:
    def __init__(self, val):
        self.val = val
        self.cntlr = val.modelXbrl.modelManager.cntlr
        self.modelXbrl = val.modelXbrl
        self.standardTaxonomiesDict = val.disclosureSystem.standardTaxonomiesDict
        self.eqCntx = {} # contexts which are OIM-aspect equivalent
        self.eqUnit = {} # units which are equivalent
        self.sumInitArcrole = self.perBindArcrole = self.aggBindArcrole = None
        self.sumConceptBindKeys = defaultdict(set)
        self.sumBoundFacts = defaultdict(list)
        self.perConceptBindKeys = defaultdict(set)
        self.perBoundFacts = defaultdict(list)
        self.durationPeriodStarts = defaultdict(set)
        self.aggConceptBindKeys = defaultdict(set)
        self.aggBoundFacts = defaultdict(list)
        self.aggBoundConceptFacts = defaultdict(list)
        self.aggDimInit = set()

    def validate(self):
        modelXbrl = self.modelXbrl
        if not modelXbrl.contexts or not modelXbrl.facts:
            return # skip if no contexts or facts

        startedAt = time.time()

        # check balance attributes and weights, same as XBRL 2.1
        for rel in modelXbrl.relationshipSet(calc2Arcroles).modelRelationships:
            weight = rel.weight
            fromConcept = rel.fromModelObject
            toConcept = rel.toModelObject
            if fromConcept is not None and toConcept is not None:
                if rel.arcrole == aggregationDomain:
                    rel.dimension = rel.arcElement.prefixedNameQname(rel.get("dimension"))
                    if rel.dimension is None or not modelXbrl.qnameConcepts[rel.dimension].isDimensionItem:
                        modelXbrl.error("calc2e:invalidAggregationDimension",
                            _("Aggregation-domain relationship has invalid dimension %(dimension)s in link role %(linkrole)s"),
                            modelObject=rel,
                            dimension=rel.get("dimension"), linkrole=ELR)
                    elif fromConcept != toConcept or not fromConcept.isDomainMember:
                        modelXbrl.error("calc2e:invalidAggregationDomain",
                            _("Calculation relationship has invalid domain %(domain)s in link role %(linkrole)s"),
                            modelObject=rel,
                            domain=fromConcept, linkrole=ELR)
                    continue
                if rel.arcrole == balanceChanges:
                    if fromConcept.periodType != "instant" or toConcept.periodType != "duration":
                        modelXbrl.error("calc2e:invalidBalanceChangesPeriodType",
                            _("Balance-changes relationship must have instant source concept and duration target concept in link role %(linkrole)s"),
                            modelObject=rel, linkrole=ELR)
                if weight not in (1, -1):
                    modelXbrl.error("calc2e:invalidWeight",
                        _("Calculation relationship has invalid weight from %(source)s to %(target)s in link role %(linkrole)s"),
                        modelObject=rel,
                        source=fromConcept.qname, target=toConcept.qname, linkrole=ELR)
                fromBalance = fromConcept.balance
                toBalance = toConcept.balance
                if fromBalance and toBalance:
                    if (fromBalance == toBalance and weight < 0) or \
                       (fromBalance != toBalance and weight > 0):
                        modelXbrl.error("calc2e:balanceCalcWeightIllegal" +
                                        ("Negative" if weight < 0 else "Positive"),
                            _("Calculation relationship has illegal weight %(weight)s from %(source)s, %(sourceBalance)s, to %(target)s, %(targetBalance)s, in link role %(linkrole)s (per 5.1.1.2 Table 6)"),
                            modelObject=rel, weight=weight,
                            source=fromConcept.qname, target=toConcept.qname, linkrole=rel.linkrole,
                            sourceBalance=fromBalance, targetBalance=toBalance,
                            messageCodes=("calc2e:balanceCalcWeightIllegalNegative", "calc2:balanceCalcWeightIllegalPositive"))
                if not fromConcept.isNumeric or not toConcept.isNumeric:
                    modelXbrl.error("calc2e:nonNumericCalc",
                        _("Calculation relationship has illegal concept from %(source)s%(sourceNumericDecorator)s to %(target)s%(targetNumericDecorator)s in link role %(linkrole)s"),
                        modelObject=rel,
                        source=fromConcept.qname, target=toConcept.qname, linkrole=rel.linkrole,
                        sourceNumericDecorator="" if fromConcept.isNumeric else _(" (non-numeric)"),
                        targetNumericDecorator="" if toConcept.isNumeric else _(" (non-numeric)"))

        # identify equal contexts
        uniqueCntxHashes = {}
        self.modelXbrl.profileActivity()
        for cntx in modelXbrl.contexts.values():
            h = hash( (cntx.periodHash, cntx.entityIdentifierHash, cntx.dimsHash) ) # OIM-compatible hash
            if h in uniqueCntxHashes:
                if cntx.isEqualTo(uniqueCntxHashes[h]):
                    self.eqCntx[cntx] = uniqueCntxHashes[h]
            else:
                uniqueCntxHashes[h] = cntx
        del uniqueCntxHashes
        self.modelXbrl.profileActivity("... identify aspect equal contexts", minTimeToShow=1.0)

        # identify equal units
        uniqueUnitHashes = {}
        for unit in self.modelXbrl.units.values():
            h = unit.hash
            if h in uniqueUnitHashes:
                if unit.isEqualTo(uniqueUnitHashes[h]):
                    self.eqUnit[unit] = uniqueUnitHashes[h]
            else:
                uniqueUnitHashes[h] = unit
        del uniqueUnitHashes
        self.modelXbrl.profileActivity("... identify equal units", minTimeToShow=1.0)


        sectObjs = sorted(set(rel.fromModelObject # only have numerics with context and unit
                              for rel in modelXbrl.relationshipSet(sectionFact).modelRelationships
                              if rel.fromModelObject is not None and rel.fromModelObject.concept is not None),
                          key=lambda s: (s.concept.label(), s.objectIndex))  # sort into document order for consistent error messages
        if not sectObjs:
            self.modelXbrl.error("calc2e:noSections",
                            "Instance contains no sections, nothing to validate.",
                            modelObject=modelXbrl)

        # check by section
        factByConceptCntxUnit = OrderedDefaultDict(list)  # sort into document order for consistent error messages
        self.sectionFacts = []
        for sectObj in sectObjs:
            #print ("section {}".format(sectObj.concept.label()))
            self.section = sectObj.concept.label()
            sectLinkRoles = tuple(sectObj.concept.get(calc2linkroles,"").split())
            factByConceptCntxUnit.clear()
            for f in sorted((rel.toModelObject # sort into document order for consistent error messages
                             for rel in modelXbrl.relationshipSet(sectionFact,sectLinkRoles).fromModelObject(sectObj)
                             if rel.toModelObject is not None and # numeric facts with context and unit
                                rel.fromModelObject is not None and
                                rel.toModelObject.concept is not None and
                                rel.toModelObject.context is not None and
                                rel.toModelObject.unit is not None),
                            key=lambda f: f.objectIndex):
                factByConceptCntxUnit[f.qname, self.eqCntx.get(f.context,f.context), self.eqUnit.get(f.unit,f.unit)].append(f)
            for fList in factByConceptCntxUnit.values():
                f0 = fList[0]
                if len(fList) == 1:
                    self.sectionFacts.append(f0)
                else:
                    if any(f.isNil for f in fList):
                        _inConsistent = not all(f.isNil for f in fList)
                        if _inConsistent: # pick a nil fact for f0 for calc validation
                            for f in fList:
                                if f.isNil:
                                    f0 = f
                                    break
                    elif all(inferredDecimals(f) == inferredDecimals(f0) for f in fList[1:]): # same decimals
                        v0 = intervalValue(f0)
                        _inConsistent = not all(intervalValue(f) == v0 for f in fList[1:])
                    else: # not all have same decimals
                        d0 = inferredDecimals(f0)
                        aMax, bMin, _inclA, _inclB = intervalValue(f0, d0)
                        for f in fList[1:]:
                            df = inferredDecimals(f0)
                            a, b, _inclA, _inclB = intervalValue(f, df)
                            if a > aMax: aMax = a
                            if b < bMin: bMin = b
                            if df > d0: # take most accurate fact in section
                                f0 = f
                                d0 = df
                        _inConsistent = (bMin < aMax)
                    if _inConsistent:
                        modelXbrl.error("calc2e:inconsistentDuplicateInSection",
                            "Section %(section)s contained %(fact)s inconsistent in contexts equivalent to %(contextID)s: values %(values)s",
                            modelObject=fList, section=sectObj.concept.label(), fact=f0.qname, contextID=f0.contextID, values=", ".join(strTruncate(f.value, 128) for f in fList))
                    self.sectionFacts.append(f0)
            # sectionFacts now in document order and deduplicated
            #print("section {} facts {}".format(sectObj.concept.label(), ", ".join(str(f.qname)+"="+f.value for f in self.sectionFacts)))

            # depth-first calc tree
            sectCalc2RelSet = modelXbrl.relationshipSet(calc2Arcroles, sectLinkRoles)

            # indexers for section based on calc2 arcrole
            self.sumInit = False
            self.sumConceptBindKeys.clear()
            self.sumBoundFacts.clear()
            self.perInit = False
            self.perConceptBindKeys.clear()
            self.perBoundFacts.clear()
            self.durationPeriodStarts.clear()
            self.aggDimInit = set()
            self.aggConceptBindKeys.clear()
            self.aggBoundFacts.clear()
            self.aggBoundConceptFacts.clear()
            self.aggDimInit.clear()

            inferredValues = {}
            for rootConcept in sorted(sectCalc2RelSet.rootConcepts,
                                      key=lambda r: sectCalc2RelSet.fromModelObject(r)[0].order):
                self.sectTreeRel(rootConcept, 1, sectCalc2RelSet, inferredValues, {rootConcept, None})

    # recursive depth-first tree descender, returns sum
    def sectTreeRel(self, parentConcept, n, sectCalc2RelSet, inferredParentValues, visited, dimQN=None):
        childRels = sectCalc2RelSet.fromModelObject(parentConcept)
        if childRels:
            visited.add(parentConcept)
            inferredChildValues = {}

            # setup summation bind keys for child objects
            sumParentBindKeys = self.sumConceptBindKeys[parentConcept]
            boundSumKeys = set() # these are contributing fact keys, parent may be inferred
            boundSums = defaultdict(intervalZero)
            boundSummationItems = defaultdict(list)
            boundPerKeys = set()
            boundPers = defaultdict(intervalZero)
            boundDurationItems = defaultdict(list)
            boundAggKeys = set()
            boundAggs = defaultdict(intervalZero)
            boundAggItems = defaultdict(list)
            boundAggConcepts = defaultdict(set)
            for rel in childRels:
                childConcept = rel.toModelObject
                if childConcept not in visited:
                    if rel.arcrole == summationItem:
                        if not self.sumInit:
                            self.sumBindFacts()
                        boundSumKeys |= self.sumConceptBindKeys[childConcept]
                    elif rel.arcrole == balanceChanges:
                        if not self.perInit:
                            self.perBindFacts()
                        boundPerKeys |= self.perConceptBindKeys[childConcept] # these are only duration items
                    elif rel.arcrole == domainMember:
                        boundAggKeys |= self.aggConceptBindKeys[dimQN]
                        domQN = parentConcept.qname
                elif rel.arcrole == aggregationDomain: # this is in visited
                    dimQN = rel.arcElement.prefixedNameQname(rel.get("dimension"))
                    if dimQN not in self.aggDimInit:
                        self.aggBindFacts(dimQN) # bind each referenced dimension's contexts

            # depth-first descent calc tree and process item after descent
            for rel in childRels:
                childConcept = rel.toModelObject
                if childConcept not in visited:
                    # depth-first descent
                    self.sectTreeRel(childConcept, n+1, sectCalc2RelSet, inferredChildValues,  visited, dimQN)
                    # post-descent summation (allows use of inferred value)
                    if rel.arcrole == summationItem:
                        weight = rel.weightDecimal
                        for sumKey in boundSumKeys:
                            cntx, unit = sumKey
                            factKey = (childConcept, cntx, unit)
                            if factKey in self.sumBoundFacts:
                                for f in self.sumBoundFacts[factKey]:
                                    addInterval(boundSums, sumKey, intervalValue(f), weight)
                                    boundSummationItems[sumKey].append(f)
                            elif factKey in inferredChildValues:
                                addInterval(boundSums, sumKey, inferredChildValues[factKey], weight)
                            elif factKey in inferredParentValues:
                                addInterval(boundSums, sumKey, inferredParentValues[factKey], weight)
                    elif rel.arcrole == balanceChanges:
                        weight = rel.weightDecimal
                        for perKey in boundPerKeys:
                            hCntx, unit, start, end = perKey
                            factKey = (childConcept, hCntx, unit, start, end)
                            if factKey in self.perBoundFacts:
                                for f in self.perBoundFacts[factKey]:
                                    addInterval(boundPers, perKey, intervalValue(f), weight)
                                    boundDurationItems[perKey].append(f)
                            elif factKey in inferredChildValues:
                                addInterval(boundPers, perKey, inferredChildValues[factKey], weight)
                            elif factKey in inferredParentValues:
                                addInterval(boundPers, perKey, inferredParentValues[factKey], weight)
                    elif rel.arcrole == domainMember:
                        memQN = childConcept.qname
                        for aggKey in boundAggKeys:
                            hCntx, unit = aggKey
                            dimMemKey = (hCntx, unit, dimQN, memQN)
                            if dimMemKey in self.aggBoundFacts:
                                for f in self.aggBoundFacts[dimMemKey]:
                                    a, b, _inclA, _inclB = intervalValue(f)
                                    factDomKey = (f.concept, hCntx, unit, dimQN, domQN)
                                    addInterval(boundAggs, factDomKey, intervalValue(f))
                                    boundAggItems[aggKey].append(f)
                                    boundAggConcepts[aggKey].add(f.concept)
                elif rel.arcrole == aggregationDomain: # this is in visited
                    childRelSet = self.modelXbrl.relationshipSet(domainMember,rel.get("targetRole"))
                    self.sectTreeRel(childConcept, n+1, childRelSet, inferredParentValues, {None}, dimQN) # infer global to section

            # process child items bound to this calc subtree
            for sumKey in boundSumKeys:
                cntx, unit = sumKey
                factKey = (parentConcept, cntx, unit)
                ia, ib = boundSums[sumKey]
                if factKey in self.sumBoundFacts:
                    for f in self.sumBoundFacts[factKey]:
                        d = inferredDecimals(f)
                        sa, sb, _inclA, _inclB = intervalValue(f, d)
                        if ((ia is NIL) ^ (sa is NIL)) or ((ia is not NIL) and (sb < ia or sa > ib)):
                            self.modelXbrl.log('INCONSISTENCY', "calc2e:summationInconsistency",
                                _("Summation inconsistent from %(concept)s in section %(section)s reported sum %(reportedSum)s, computed sum %(computedSum)s context %(contextID)s unit %(unitID)s unreported contributing items %(unreportedContributors)s"),
                                modelObject=boundSummationItems[sumKey],
                                concept=parentConcept.qname, section=self.section,
                                reportedSum=self.formatInterval(sa, sb, d),
                                computedSum=self.formatInterval(ia, ib, d),
                                contextID=f.context.id, unitID=f.unit.id,
                                unreportedContributors=", ".join(str(c.qname) # list the missing/unreported contributors in relationship order
                                                                 for r in childRels
                                                                 for c in (r.toModelObject,)
                                                                 if r.arcrole == summationItem and c is not None and
                                                                 (c, cntx, unit) not in self.sumBoundFacts)
                                                         or "none")
                elif inferredParentValues is not None: # value was inferred, return to parent level
                    inferredParentValues[factKey] = (ia, ib)
            for perKey in boundPerKeys:
                hCntx, unit, start, end = perKey
                ia, ib = boundPers[perKey]
                endBalA = endBalB = ZERO
                endFactKey = (parentConcept, hCntx, unit, None, end)
                if endFactKey in self.perBoundFacts:
                    for f in self.perBoundFacts[endFactKey]:
                        if f.isNil:
                            endBalA = endBalB = NIL
                            d = 0
                            break
                        d = inferredDecimals(f)
                        a, b, _inclA, _inclB = intervalValue(f,d)
                        endBalA += a
                        endBalB += b
                    foundStartingFact = (endBalA is NIL)
                    while not foundStartingFact:
                        startFactKey = (parentConcept, hCntx, unit, None, start)
                        if startFactKey in self.perBoundFacts:
                            for f in self.perBoundFacts[startFactKey]:
                                if f.isNil:
                                    endBalA = endBalB = NIL
                                    foundStartingFact = True
                                    break
                                a, b, _inclA, _inclB = intervalValue(f)
                                endBalA -= a
                                endBalB -= b
                                foundStartingFact = True
                                break
                        if not foundStartingFact:
                            # infer backing up one period
                            _nomPer = nominalPeriod(end - start)
                            foundEarlierAdjacentPeriodStart = False
                            for _start in self.durationPeriodStarts.get(_nomPer, ()):
                                if nominalPeriod(start - _start) == _nomPer: # it's preceding period
                                    end = start
                                    start = _start
                                    perKey = hCntx, unit, start, end
                                    if perKey in boundPerKeys:
                                        chngs = boundPers[perKey]
                                        ia += chngs[0]
                                        ib += chngs[1]
                                        foundEarlierAdjacentPeriodStart = True
                                        break
                            if not foundEarlierAdjacentPeriodStart:
                                break

                    if ((ia is NIL) ^ (endBalA is NIL)) or ((ia is not NIL) and (endBalB < ia or endBalA > ib)):
                        self.modelXbrl.log('INCONSISTENCY', "calc2e:balanceInconsistency",
                            _("Balance inconsistent from %(concept)s in section %(section)s reported sum %(reportedSum)s, computed sum %(computedSum)s context %(contextID)s unit %(unitID)s unreported contributing items %(unreportedContributors)s"),
                            modelObject=boundDurationItems[perKey],
                            concept=parentConcept.qname, section=self.section,
                            reportedSum=self.formatInterval(endBalA, endBalB, d),
                            computedSum=self.formatInterval(ia, ib, d),
                            contextID=f.context.id, unitID=f.unit.id,
                            unreportedContributors=", ".join(str(c.qname) # list the missing/unreported contributors in relationship order
                                                             for r in childRels
                                                             for c in (r.toModelObject,)
                                                             if r.arcrole == balanceChanges and c is not None and
                                                             (c, hCntx, unit, start, end) not in self.perBoundFacts)
                                                     or "none")
            for aggKey in boundAggKeys:
                hCntx, unit = aggKey
                for concept in sorted(boundAggConcepts[aggKey], key=lambda c:c.objectIndex): # repeatable errors
                    factDomKey = (concept, hCntx, unit, dimQN, domQN)
                    ia, ib = boundAggs[factDomKey]
                    if factDomKey in self.aggBoundConceptFacts:
                        for f in self.aggBoundConceptFacts[factDomKey]:
                            d = inferredDecimals(f)
                            sa, sb, _inclA, _inclB = intervalValue(f, d)
                            if ((ia is NIL) ^ (sa is NIL)) or ((ia is not NIL) and (sb < ia or sa > ib)):
                                self.modelXbrl.log('INCONSISTENCY', "calc2e:aggregationInconsistency",
                                    _("Aggregation inconsistent for %(concept)s, domain %(domain)s in section %(section)s reported sum %(reportedSum)s, computed sum %(computedSum)s context %(contextID)s unit %(unitID)s unreported contributing members %(unreportedContributors)s"),
                                    modelObject=boundAggItems[factDomKey],
                                    concept=concept.qname,
                                    domain=parentConcept.qname, section=self.section,
                                    reportedSum=self.formatInterval(sa, sb, d),
                                    computedSum=self.formatInterval(ia, ib, d),
                                    contextID=f.context.id, unitID=f.unit.id,
                                    unreportedContributors=", ".join(str(c.qname) # list the missing/unreported contributors in relationship order
                                                                     for r in childRels
                                                                     for c in (r.toModelObject,)
                                                                     if r.arcrole == domainMember and c is not None and
                                                                     (concept, hCntx, unit, dimQN, c.qname) not in self.aggBoundConceptFacts)
                                                             or "none")
                    elif inferredParentValues is not None: # value was inferred, return to parent level
                        # allow to be retrieved by factDomKey
                        inferredParentValues[factDomKey] = (ia, ib)
                        if self.modelXbrl.qnameDimensionDefaults.get(dimQN) == domQN:
                            cntxKey = (hCntx, dimQN, domQN)
                            if cntxKey in self.eqCntx:
                                cntx = self.eqCntx[cntxKey]
                            else:
                                cntx = self.aggTotalContext(hCntx, dimQN, domQN)
                                self.eqCntx[cntxKey] = cntx
                            if cntx is not None:
                                # allow to be retrieved by fact line item context key
                                self.eqCntx[(hCntx, dimQN, domQN)] = cntx
                                inferredParentValues[(concept, cntx, unit)] = (ia, ib)
            visited.remove(parentConcept)

    def sumBindFacts(self):
        # bind facts in section for summation-item
        for f in self.sectionFacts:
            concept = f.concept
            if concept.isNumeric:
                cntx = self.eqCntx.get(f.context,f.context)
                unit = self.eqUnit.get(f.unit,f.unit)
                self.sumConceptBindKeys[concept].add( (cntx,unit) )
                self.sumBoundFacts[concept, cntx, unit].append(f)
        self.sumInit = True

    def perBindFacts(self):
        # bind facts in section for domain aggreggation
        for f in self.sectionFacts:
            concept = f.concept
            if concept.isNumeric:
                cntx = self.eqCntx.get(f.context,f.context)
                if not cntx.isForeverPeriod:
                    hCntx = hash( (cntx.entityIdentifierHash, cntx.dimsHash) )
                    unit = self.eqUnit.get(f.unit,f.unit)
                    self.perConceptBindKeys[concept].add( (hCntx, unit, cntx.startDatetime, cntx.endDatetime) )
                    self.perBoundFacts[concept, hCntx, unit, cntx.startDatetime, cntx.endDatetime].append(f)
                    if cntx.isStartEndPeriod:
                        self.durationPeriodStarts[nominalPeriod(cntx.endDatetime - cntx.startDatetime)].add(cntx.startDatetime)
        self.perInit = True

    def aggBindFacts(self, dimQN):
        # bind facts in section for domain aggreggation
        for f in self.sectionFacts:
            concept = f.concept
            if concept.isNumeric:
                cntx = self.eqCntx.get(f.context,f.context)
                hCntx = hash( (cntx.periodHash, cntx.entityIdentifierHash,
                               hash(frozenset(dimObj
                                              for _dimQN, dimObj in cntx.qnameDims.items()
                                            if _dimQN != dimQN))) )
                unit = self.eqUnit.get(f.unit,f.unit)
                memQN = cntx.dimMemberQname(dimQN, includeDefaults=True)
                if memQN is not None:
                    self.aggConceptBindKeys[dimQN].add( (hCntx,unit) )
                    self.aggBoundFacts[hCntx, unit, dimQN, memQN].append(f)
                    self.aggBoundConceptFacts[f.concept, hCntx, unit, dimQN, memQN].append(f)
        self.aggDimInit.add(dimQN)

    def aggTotalContext(self, hCntx, dimQN, domQN):
        # find a context for aggregation total usable in line item and balance roll ups
        for cntx in self.modelXbrl.contexts.values():
            hCntx2 = hash( (cntx.periodHash, cntx.entityIdentifierHash,
                            hash(frozenset(dimObj
                                           for _dimQN, dimObj in cntx.qnameDims.items()
                                           if _dimQN != dimQN))) )
            if hCntx == hCntx2 and cntx.dimMemberQname(dimQN, True) == domQN:
                return cntx
        return None

    def formatInterval(self, a, b, dec):
        if a is NIL:
            return "(nil)"
        if isnan(dec) or isinf(dec): dec = 4
        if a == b: # not an interval
            return Locale.format_decimal(self.modelXbrl.locale, a, 1, max(dec,0))
        return "[{}, {}]".format( # show as an interval
            Locale.format_decimal(self.modelXbrl.locale, a, 1, max(dec,0)),
            Locale.format_decimal(self.modelXbrl.locale, b, 1, max(dec,0)))


def checkCalc2(val, *args, **kwargs):
    ValidateXbrlCalc2(val).validate()



__pluginInfo__ = {
    # Do not use _( ) in pluginInfo itself (it is applied later, after loading
    'name': 'Calc2',
    'version': '0.9',
    'description': '''Calculation 2.0 Validation.''',
    'license': 'Apache-2',
    'author': authorLabel,
    'copyright': copyrightLabel,
    # classes of mount points (required)
    'Validate.XBRL.Finally': checkCalc2
}



from pathlib import PurePath
from tests.integration_tests.validation.conformance_suite_config import ConformanceSuiteConfig

config = ConformanceSuiteConfig(
    args=[
        '--formula', 'run',
    ],
    file='testcase.xml',
    info_url='https://specifications.xbrl.org/work-product-index-inline-xbrl-transformation-registry-3.html',
    local_filepath='trr-3.0.1.zip',
    membership_url='https://www.xbrl.org/join',
    name=PurePath(__file__).stem,
)



from pathlib import PurePath
from tests.integration_tests.validation.conformance_suite_config import ConformanceSuiteConfig

config = ConformanceSuiteConfig(
    args=[
        '--formula', 'run',
    ],
    expected_failure_ids=frozenset(f'table-linkbase-conf-2015-08-12/conf/tests/{s}' for s in [
        '0100-table/0144-table-filter-arc/0144-table-filter-arc-testcase.xml:v-03',
        '0100-table/0144-table-filter-arc/0144-table-filter-arc-testcase.xml:v-04',
        '0100-table/0144-table-filter-arc/0144-table-filter-arc-testcase.xml:v-05',
        '0100-table/0150-breakdown-without-aspects/0150-breakdown-without-aspects-testcase.xml:v-04',
        '0100-table/0150-breakdown-without-aspects/0150-breakdown-without-aspects-testcase.xml:v-04i',
        '0100-table/0160-non-participating-aspects/0160-non-participating-aspects-testcase.xml:v-03',
        '0100-table/0170-parent-child-order/0170-parent-child-order-testcase.xml:v-01',
        '0200-table-parameters/0200-table-parameters-testcase.xml:v-02i',
        '1000-rule-node/1000-rule-node-testcase.xml:v-03',
        '1000-rule-node/1000-rule-node-testcase.xml:v-03i',
        '1000-rule-node/1000-rule-node-testcase.xml:v-04',
        '1000-rule-node/1000-rule-node-testcase.xml:v-04i',
        '1000-rule-node/1000-rule-node-testcase.xml:v-05',
        '1000-rule-node/1000-rule-node-testcase.xml:v-06',
        '1000-rule-node/1000-rule-node-testcase.xml:v-07',
        '1000-rule-node/1000-rule-node-testcase.xml:v-07i',
        '1000-rule-node/1000-rule-node-testcase.xml:v-07n',
        '1000-rule-node/1000-rule-node-testcase.xml:v-07ni',
        '1000-rule-node/1000-rule-node-testcase.xml:v-08',
        '1000-rule-node/1010-rule-node-concept-rule/1010-rule-node-concept-rule-testcase.xml:v-05',
        '1000-rule-node/1020-rule-node-explicit-dimension-rule/1020-rule-node-explicit-dimension-rule-testcase.xml:v-01',
        '1000-rule-node/1020-rule-node-explicit-dimension-rule/1020-rule-node-explicit-dimension-rule-testcase.xml:v-01i',
        '1000-rule-node/1020-rule-node-explicit-dimension-rule/1020-rule-node-explicit-dimension-rule-testcase.xml:v-02',
        '1000-rule-node/1020-rule-node-explicit-dimension-rule/1020-rule-node-explicit-dimension-rule-testcase.xml:v-02i',
        '1000-rule-node/1020-rule-node-explicit-dimension-rule/1020-rule-node-explicit-dimension-rule-testcase.xml:v-03',
        '1000-rule-node/1020-rule-node-explicit-dimension-rule/1020-rule-node-explicit-dimension-rule-testcase.xml:v-04',
        '1000-rule-node/1020-rule-node-explicit-dimension-rule/1020-rule-node-explicit-dimension-rule-testcase.xml:v-04i',
        '1000-rule-node/1020-rule-node-explicit-dimension-rule/1020-rule-node-explicit-dimension-rule-testcase.xml:v-06',
        '1000-rule-node/1020-rule-node-explicit-dimension-rule/1020-rule-node-explicit-dimension-rule-testcase.xml:v-06i',
        '1000-rule-node/1020-rule-node-explicit-dimension-rule/1020-rule-node-explicit-dimension-rule-testcase.xml:v-07',
        '1000-rule-node/1020-rule-node-explicit-dimension-rule/1020-rule-node-explicit-dimension-rule-testcase.xml:v-07i',
        '1000-rule-node/1030-rule-node-typed-dimension-rule/1030-rule-node-typed-dimension-rule-testcase.xml:v-01',
        '1000-rule-node/1030-rule-node-typed-dimension-rule/1030-rule-node-typed-dimension-rule-testcase.xml:v-01i',
        '1000-rule-node/1030-rule-node-typed-dimension-rule/1030-rule-node-typed-dimension-rule-testcase.xml:v-02',
        '1000-rule-node/1030-rule-node-typed-dimension-rule/1030-rule-node-typed-dimension-rule-testcase.xml:v-02i',
        '1000-rule-node/1030-rule-node-typed-dimension-rule/1030-rule-node-typed-dimension-rule-testcase.xml:v-03',
        '1000-rule-node/1030-rule-node-typed-dimension-rule/1030-rule-node-typed-dimension-rule-testcase.xml:v-03i',
        '1000-rule-node/1040-rule-node-period-rule/1040-rule-node-period-rule-testcase.xml:v-01',
        '1000-rule-node/1040-rule-node-period-rule/1040-rule-node-period-rule-testcase.xml:v-01i',
        '1000-rule-node/1040-rule-node-period-rule/1040-rule-node-period-rule-testcase.xml:v-02',
        '1000-rule-node/1040-rule-node-period-rule/1040-rule-node-period-rule-testcase.xml:v-02i',
        '1000-rule-node/1040-rule-node-period-rule/1040-rule-node-period-rule-testcase.xml:v-03',
        '1000-rule-node/1040-rule-node-period-rule/1040-rule-node-period-rule-testcase.xml:v-03i',
        '1000-rule-node/1040-rule-node-period-rule/1040-rule-node-period-rule-testcase.xml:v-04',
        '1000-rule-node/1050-rule-node-entity-rule/1050-rule-node-entity-rule-testcase.xml:v-01',
        '1000-rule-node/1050-rule-node-entity-rule/1050-rule-node-entity-rule-testcase.xml:v-01i',
        '1000-rule-node/1060-rule-node-unit-rule/1060-rule-node-unit-rule-testcase.xml:v-01',
        '1000-rule-node/1060-rule-node-unit-rule/1060-rule-node-unit-rule-testcase.xml:v-01i',
        '1000-rule-node/1070-rule-node-occ-rule/1070-rule-node-occ-rule-testcase.xml:v-01',
        '1000-rule-node/1070-rule-node-occ-rule/1070-rule-node-occ-rule-testcase.xml:v-01i',
        '1000-rule-node/1070-rule-node-occ-rule/1070-rule-node-occ-rule-testcase.xml:v-02',
        '1000-rule-node/1070-rule-node-occ-rule/1070-rule-node-occ-rule-testcase.xml:v-02i',
        '1000-rule-node/1070-rule-node-occ-rule/1070-rule-node-occ-rule-testcase.xml:v-03i',
        '1000-rule-node/1070-rule-node-occ-rule/1070-rule-node-occ-rule-testcase.xml:v-04i',
        '1000-rule-node/1070-rule-node-occ-rule/1070-rule-node-occ-rule-testcase.xml:v-05',
        '1000-rule-node/1070-rule-node-occ-rule/1070-rule-node-occ-rule-testcase.xml:v-05i',
        '1000-rule-node/1070-rule-node-occ-rule/1070-rule-node-occ-rule-testcase.xml:v-06',
        '1000-rule-node/1070-rule-node-occ-rule/1070-rule-node-occ-rule-testcase.xml:v-06i',
        '1000-rule-node/1070-rule-node-occ-rule/1070-rule-node-occ-rule-testcase.xml:v-07',
        '1000-rule-node/1070-rule-node-occ-rule/1070-rule-node-occ-rule-testcase.xml:v-07i',
        '1000-rule-node/1100-abstract-rule-node/1100-abstract-rule-node-testcase.xml:v-01',
        '1000-rule-node/1100-abstract-rule-node/1100-abstract-rule-node-testcase.xml:v-01i',
        '1000-rule-node/1200-merged-rule-node/1200-merged-rule-node-testcase.xml:v-01',
        '1000-rule-node/1200-merged-rule-node/1200-merged-rule-node-testcase.xml:v-01i',
        '1000-rule-node/1200-merged-rule-node/1200-merged-rule-node-testcase.xml:v-02',
        '1000-rule-node/1200-merged-rule-node/1200-merged-rule-node-testcase.xml:v-02i',
        '1000-rule-node/1200-merged-rule-node/1200-merged-rule-node-testcase.xml:v-03',
        '1000-rule-node/1200-merged-rule-node/1200-merged-rule-node-testcase.xml:v-03i',
        '1000-rule-node/1200-merged-rule-node/1200-merged-rule-node-testcase.xml:v-04',
        '1000-rule-node/1200-merged-rule-node/1200-merged-rule-node-testcase.xml:v-04i',
        '1000-rule-node/1200-merged-rule-node/1200-merged-rule-node-testcase.xml:v-05',
        '1000-rule-node/1200-merged-rule-node/1200-merged-rule-node-testcase.xml:v-05i',
        '1000-rule-node/1200-merged-rule-node/1200-merged-rule-node-testcase.xml:v-09',
        '1000-rule-node/1200-merged-rule-node/1200-merged-rule-node-testcase.xml:v-09i',
        '1000-rule-node/1200-merged-rule-node/1200-merged-rule-node-testcase.xml:v-10',
        '1000-rule-node/1200-merged-rule-node/1200-merged-rule-node-testcase.xml:v-10i',
        '1000-rule-node/1200-merged-rule-node/1200-merged-rule-node-testcase.xml:v-11',
        '1000-rule-node/1200-merged-rule-node/1200-merged-rule-node-testcase.xml:v-11i',
        '1000-rule-node/1200-merged-rule-node/1200-merged-rule-node-testcase.xml:v-12',
        '1000-rule-node/1200-merged-rule-node/1200-merged-rule-node-testcase.xml:v-12i',
        '1000-rule-node/1300-rule-node-rule-sets/1300-rule-node-rule-sets-testcase.xml:v-01',
        '1000-rule-node/1300-rule-node-rule-sets/1300-rule-node-rule-sets-testcase.xml:v-01i',
        '1000-rule-node/1300-rule-node-rule-sets/1300-rule-node-rule-sets-testcase.xml:v-06',
        '1000-rule-node/1300-rule-node-rule-sets/1300-rule-node-rule-sets-testcase.xml:v-06i',
        '1000-rule-node/1300-rule-node-rule-sets/1300-rule-node-rule-sets-testcase.xml:v-07',
        '1000-rule-node/1300-rule-node-rule-sets/1300-rule-node-rule-sets-testcase.xml:v-07i',
        '1000-rule-node/1300-rule-node-rule-sets/1300-rule-node-rule-sets-testcase.xml:v-08',
        '1000-rule-node/1300-rule-node-rule-sets/1300-rule-node-rule-sets-testcase.xml:v-08i',
        '1000-rule-node/1300-rule-node-rule-sets/1300-rule-node-rule-sets-testcase.xml:v-09',
        '1000-rule-node/1300-rule-node-rule-sets/1300-rule-node-rule-sets-testcase.xml:v-10',
        '3100-concept-relationship-node/3100-concept-relationship-node-testcase.xml:v-01',
        '3100-concept-relationship-node/3100-concept-relationship-node-testcase.xml:v-02',
        '3100-concept-relationship-node/3110-concept-relationship-node-relationship-source/3110-concept-relationship-node-relationship-source-testcase.xml:v-01',
        '3100-concept-relationship-node/3110-concept-relationship-node-relationship-source/3110-concept-relationship-node-relationship-source-testcase.xml:v-01i',
        '3100-concept-relationship-node/3110-concept-relationship-node-relationship-source/3110-concept-relationship-node-relationship-source-testcase.xml:v-02',
        '3100-concept-relationship-node/3110-concept-relationship-node-relationship-source/3110-concept-relationship-node-relationship-source-testcase.xml:v-02i',
        '3100-concept-relationship-node/3110-concept-relationship-node-relationship-source/3110-concept-relationship-node-relationship-source-testcase.xml:v-03',
        '3100-concept-relationship-node/3110-concept-relationship-node-relationship-source/3110-concept-relationship-node-relationship-source-testcase.xml:v-03i',
        '3100-concept-relationship-node/3110-concept-relationship-node-relationship-source/3110-concept-relationship-node-relationship-source-testcase.xml:v-04',
        '3100-concept-relationship-node/3110-concept-relationship-node-relationship-source/3110-concept-relationship-node-relationship-source-testcase.xml:v-04i',
        '3100-concept-relationship-node/3110-concept-relationship-node-relationship-source/3110-concept-relationship-node-relationship-source-testcase.xml:v-06',
        '3100-concept-relationship-node/3110-concept-relationship-node-relationship-source/3110-concept-relationship-node-relationship-source-testcase.xml:v-07',
        '3100-concept-relationship-node/3110-concept-relationship-node-relationship-source/3110-concept-relationship-node-relationship-source-testcase.xml:v-08',
        '3100-concept-relationship-node/3110-concept-relationship-node-relationship-source/3110-concept-relationship-node-relationship-source-testcase.xml:v-08i',
        '3100-concept-relationship-node/3110-concept-relationship-node-relationship-source/3110-concept-relationship-node-relationship-source-testcase.xml:v-09',
        '3100-concept-relationship-node/3110-concept-relationship-node-relationship-source/3110-concept-relationship-node-relationship-source-testcase.xml:v-09i',
        '3100-concept-relationship-node/3120-concept-relationship-node-linkrole/3120-concept-relationship-node-linkrole-testcase.xml:v-01',
        '3100-concept-relationship-node/3120-concept-relationship-node-linkrole/3120-concept-relationship-node-linkrole-testcase.xml:v-01i',
        '3100-concept-relationship-node/3120-concept-relationship-node-linkrole/3120-concept-relationship-node-linkrole-testcase.xml:v-02',
        '3100-concept-relationship-node/3120-concept-relationship-node-linkrole/3120-concept-relationship-node-linkrole-testcase.xml:v-03',
        '3100-concept-relationship-node/3120-concept-relationship-node-linkrole/3120-concept-relationship-node-linkrole-testcase.xml:v-04',
        '3100-concept-relationship-node/3120-concept-relationship-node-linkrole/3120-concept-relationship-node-linkrole-testcase.xml:v-05',
        '3100-concept-relationship-node/3120-concept-relationship-node-linkrole/3120-concept-relationship-node-linkrole-testcase.xml:v-06',
        '3100-concept-relationship-node/3130-concept-relationship-node-arcrole/3130-concept-relationship-node-arcrole-testcase.xml:v-01',
        '3100-concept-relationship-node/3130-concept-relationship-node-arcrole/3130-concept-relationship-node-arcrole-testcase.xml:v-01i',
        '3100-concept-relationship-node/3130-concept-relationship-node-arcrole/3130-concept-relationship-node-arcrole-testcase.xml:v-02',
        '3100-concept-relationship-node/3130-concept-relationship-node-arcrole/3130-concept-relationship-node-arcrole-testcase.xml:v-02i',
        '3100-concept-relationship-node/3130-concept-relationship-node-arcrole/3130-concept-relationship-node-arcrole-testcase.xml:v-03',
        '3100-concept-relationship-node/3130-concept-relationship-node-arcrole/3130-concept-relationship-node-arcrole-testcase.xml:v-03i',
        '3100-concept-relationship-node/3130-concept-relationship-node-arcrole/3130-concept-relationship-node-arcrole-testcase.xml:v-04',
        '3100-concept-relationship-node/3130-concept-relationship-node-arcrole/3130-concept-relationship-node-arcrole-testcase.xml:v-05',
        '3100-concept-relationship-node/3130-concept-relationship-node-arcrole/3130-concept-relationship-node-arcrole-testcase.xml:v-06',
        '3100-concept-relationship-node/3130-concept-relationship-node-arcrole/3130-concept-relationship-node-arcrole-testcase.xml:v-07',
        '3100-concept-relationship-node/3140-concept-relationship-node-linkname/3140-concept-relationship-node-linkname-testcase.xml:v-01',
        '3100-concept-relationship-node/3140-concept-relationship-node-linkname/3140-concept-relationship-node-linkname-testcase.xml:v-01i',
        '3100-concept-relationship-node/3140-concept-relationship-node-linkname/3140-concept-relationship-node-linkname-testcase.xml:v-02',
        '3100-concept-relationship-node/3140-concept-relationship-node-linkname/3140-concept-relationship-node-linkname-testcase.xml:v-02i',
        '3100-concept-relationship-node/3140-concept-relationship-node-linkname/3140-concept-relationship-node-linkname-testcase.xml:v-03',
        '3100-concept-relationship-node/3140-concept-relationship-node-linkname/3140-concept-relationship-node-linkname-testcase.xml:v-03i',
        '3100-concept-relationship-node/3140-concept-relationship-node-linkname/3140-concept-relationship-node-linkname-testcase.xml:v-04',
        '3100-concept-relationship-node/3140-concept-relationship-node-linkname/3140-concept-relationship-node-linkname-testcase.xml:v-05',
        '3100-concept-relationship-node/3150-concept-relationship-node-arcname/3150-concept-relationship-node-arcname-testcase.xml:v-01',
        '3100-concept-relationship-node/3150-concept-relationship-node-arcname/3150-concept-relationship-node-arcname-testcase.xml:v-01i',
        '3100-concept-relationship-node/3150-concept-relationship-node-arcname/3150-concept-relationship-node-arcname-testcase.xml:v-02',
        '3100-concept-relationship-node/3150-concept-relationship-node-arcname/3150-concept-relationship-node-arcname-testcase.xml:v-02i',
        '3100-concept-relationship-node/3150-concept-relationship-node-arcname/3150-concept-relationship-node-arcname-testcase.xml:v-03',
        '3100-concept-relationship-node/3150-concept-relationship-node-arcname/3150-concept-relationship-node-arcname-testcase.xml:v-03i',
        '3100-concept-relationship-node/3150-concept-relationship-node-arcname/3150-concept-relationship-node-arcname-testcase.xml:v-04',
        '3100-concept-relationship-node/3150-concept-relationship-node-arcname/3150-concept-relationship-node-arcname-testcase.xml:v-05',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-01',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-01i',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-02',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-02i',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-03',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-03i',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-04',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-04i',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-05',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-05i',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-06',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-06i',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-07',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-07i',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-08',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-08i',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-08a',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-08ai',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-09',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-09i',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-10',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-10i',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-11',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-11i',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-12',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-12i',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-13',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-14',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-15',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-15i',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-16',
        '3100-concept-relationship-node/3160-concept-relationship-node-formula-axis/3160-concept-relationship-node-formula-axis-testcase.xml:v-17',
        '3100-concept-relationship-node/3170-concept-relationship-node-generations/3170-concept-relationship-node-generations-testcase.xml:v-01',
        '3100-concept-relationship-node/3170-concept-relationship-node-generations/3170-concept-relationship-node-generations-testcase.xml:v-01i',
        '3100-concept-relationship-node/3170-concept-relationship-node-generations/3170-concept-relationship-node-generations-testcase.xml:v-02',
        '3100-concept-relationship-node/3170-concept-relationship-node-generations/3170-concept-relationship-node-generations-testcase.xml:v-02i',
        '3100-concept-relationship-node/3170-concept-relationship-node-generations/3170-concept-relationship-node-generations-testcase.xml:v-03',
        '3100-concept-relationship-node/3170-concept-relationship-node-generations/3170-concept-relationship-node-generations-testcase.xml:v-03i',
        '3100-concept-relationship-node/3170-concept-relationship-node-generations/3170-concept-relationship-node-generations-testcase.xml:v-04',
        '3100-concept-relationship-node/3170-concept-relationship-node-generations/3170-concept-relationship-node-generations-testcase.xml:v-05',
        '3100-concept-relationship-node/3170-concept-relationship-node-generations/3170-concept-relationship-node-generations-testcase.xml:v-05i',
        '3100-concept-relationship-node/3170-concept-relationship-node-generations/3170-concept-relationship-node-generations-testcase.xml:v-06',
        '3100-concept-relationship-node/3170-concept-relationship-node-generations/3170-concept-relationship-node-generations-testcase.xml:v-07',
        '3100-concept-relationship-node/3170-concept-relationship-node-generations/3170-concept-relationship-node-generations-testcase.xml:v-08',
        '3100-concept-relationship-node/3180-concept-relationship-node-tag-selectors/3180-concept-relationship-node-tag-selectors-testcase.xml:v-01',
        '3100-concept-relationship-node/3180-concept-relationship-node-tag-selectors/3180-concept-relationship-node-tag-selectors-testcase.xml:v-02',
        '3100-concept-relationship-node/3180-concept-relationship-node-tag-selectors/3180-concept-relationship-node-tag-selectors-testcase.xml:v-03',
        '3200-dimension-relationship-node/3200-dimension-relationship-node-testcase.xml:v-01',
        '3200-dimension-relationship-node/3200-dimension-relationship-node-testcase.xml:v-01i',
        '3200-dimension-relationship-node/3200-dimension-relationship-node-testcase.xml:v-02',
        '3200-dimension-relationship-node/3200-dimension-relationship-node-testcase.xml:v-03',
        '3200-dimension-relationship-node/3200-dimension-relationship-node-testcase.xml:v-03i',
        '3200-dimension-relationship-node/3200-dimension-relationship-node-testcase.xml:v-04',
        '3200-dimension-relationship-node/3210-dimension-relationship-node-relationship-source/3210-dimension-relationship-node-relationship-source-testcase.xml:v-01',
        '3200-dimension-relationship-node/3210-dimension-relationship-node-relationship-source/3210-dimension-relationship-node-relationship-source-testcase.xml:v-01i',
        '3200-dimension-relationship-node/3210-dimension-relationship-node-relationship-source/3210-dimension-relationship-node-relationship-source-testcase.xml:v-02',
        '3200-dimension-relationship-node/3210-dimension-relationship-node-relationship-source/3210-dimension-relationship-node-relationship-source-testcase.xml:v-02i',
        '3200-dimension-relationship-node/3210-dimension-relationship-node-relationship-source/3210-dimension-relationship-node-relationship-source-testcase.xml:v-03',
        '3200-dimension-relationship-node/3210-dimension-relationship-node-relationship-source/3210-dimension-relationship-node-relationship-source-testcase.xml:v-03i',
        '3200-dimension-relationship-node/3210-dimension-relationship-node-relationship-source/3210-dimension-relationship-node-relationship-source-testcase.xml:v-08',
        '3200-dimension-relationship-node/3210-dimension-relationship-node-relationship-source/3210-dimension-relationship-node-relationship-source-testcase.xml:v-08i',
        '3200-dimension-relationship-node/3210-dimension-relationship-node-relationship-source/3210-dimension-relationship-node-relationship-source-testcase.xml:v-09',
        '3200-dimension-relationship-node/3210-dimension-relationship-node-relationship-source/3210-dimension-relationship-node-relationship-source-testcase.xml:v-09i',
        '3200-dimension-relationship-node/3210-dimension-relationship-node-relationship-source/3210-dimension-relationship-node-relationship-source-testcase.xml:v-10',
        '3200-dimension-relationship-node/3210-dimension-relationship-node-relationship-source/3210-dimension-relationship-node-relationship-source-testcase.xml:v-11',
        '3200-dimension-relationship-node/3220-dimension-relationship-node-linkrole/3220-dimension-relationship-node-linkrole-testcase.xml:v-01',
        '3200-dimension-relationship-node/3220-dimension-relationship-node-linkrole/3220-dimension-relationship-node-linkrole-testcase.xml:v-01i',
        '3200-dimension-relationship-node/3220-dimension-relationship-node-linkrole/3220-dimension-relationship-node-linkrole-testcase.xml:v-02',
        '3200-dimension-relationship-node/3220-dimension-relationship-node-linkrole/3220-dimension-relationship-node-linkrole-testcase.xml:v-02i',
        '3200-dimension-relationship-node/3220-dimension-relationship-node-linkrole/3220-dimension-relationship-node-linkrole-testcase.xml:v-03',
        '3200-dimension-relationship-node/3220-dimension-relationship-node-linkrole/3220-dimension-relationship-node-linkrole-testcase.xml:v-04',
        '3200-dimension-relationship-node/3220-dimension-relationship-node-linkrole/3220-dimension-relationship-node-linkrole-testcase.xml:v-05',
        '3200-dimension-relationship-node/3220-dimension-relationship-node-linkrole/3220-dimension-relationship-node-linkrole-testcase.xml:v-06',
        '3200-dimension-relationship-node/3220-dimension-relationship-node-linkrole/3220-dimension-relationship-node-linkrole-testcase.xml:v-07',
        '3200-dimension-relationship-node/3240-dimension-relationship-node-formula-axis/3240-dimension-relationship-node-formula-axis-testcase.xml:v-01',
        '3200-dimension-relationship-node/3240-dimension-relationship-node-formula-axis/3240-dimension-relationship-node-formula-axis-testcase.xml:v-01i',
        '3200-dimension-relationship-node/3240-dimension-relationship-node-formula-axis/3240-dimension-relationship-node-formula-axis-testcase.xml:v-02',
        '3200-dimension-relationship-node/3240-dimension-relationship-node-formula-axis/3240-dimension-relationship-node-formula-axis-testcase.xml:v-02i',
        '3200-dimension-relationship-node/3240-dimension-relationship-node-formula-axis/3240-dimension-relationship-node-formula-axis-testcase.xml:v-03',
        '3200-dimension-relationship-node/3240-dimension-relationship-node-formula-axis/3240-dimension-relationship-node-formula-axis-testcase.xml:v-03i',
        '3200-dimension-relationship-node/3240-dimension-relationship-node-formula-axis/3240-dimension-relationship-node-formula-axis-testcase.xml:v-04',
        '3200-dimension-relationship-node/3240-dimension-relationship-node-formula-axis/3240-dimension-relationship-node-formula-axis-testcase.xml:v-04i',
        '3200-dimension-relationship-node/3240-dimension-relationship-node-formula-axis/3240-dimension-relationship-node-formula-axis-testcase.xml:v-05',
        '3200-dimension-relationship-node/3240-dimension-relationship-node-formula-axis/3240-dimension-relationship-node-formula-axis-testcase.xml:v-05i',
        '3200-dimension-relationship-node/3240-dimension-relationship-node-formula-axis/3240-dimension-relationship-node-formula-axis-testcase.xml:v-06',
        '3200-dimension-relationship-node/3240-dimension-relationship-node-formula-axis/3240-dimension-relationship-node-formula-axis-testcase.xml:v-06i',
        '3200-dimension-relationship-node/3240-dimension-relationship-node-formula-axis/3240-dimension-relationship-node-formula-axis-testcase.xml:v-07',
        '3200-dimension-relationship-node/3240-dimension-relationship-node-formula-axis/3240-dimension-relationship-node-formula-axis-testcase.xml:v-08',
        '3200-dimension-relationship-node/3250-dimension-relationship-node-generations/3250-dimension-relationship-node-generations-testcase.xml:v-01',
        '3200-dimension-relationship-node/3250-dimension-relationship-node-generations/3250-dimension-relationship-node-generations-testcase.xml:v-01i',
        '3200-dimension-relationship-node/3250-dimension-relationship-node-generations/3250-dimension-relationship-node-generations-testcase.xml:v-02',
        '3200-dimension-relationship-node/3250-dimension-relationship-node-generations/3250-dimension-relationship-node-generations-testcase.xml:v-02i',
        '3200-dimension-relationship-node/3250-dimension-relationship-node-generations/3250-dimension-relationship-node-generations-testcase.xml:v-03',
        '3200-dimension-relationship-node/3250-dimension-relationship-node-generations/3250-dimension-relationship-node-generations-testcase.xml:v-03i',
        '3200-dimension-relationship-node/3250-dimension-relationship-node-generations/3250-dimension-relationship-node-generations-testcase.xml:v-04',
        '3200-dimension-relationship-node/3250-dimension-relationship-node-generations/3250-dimension-relationship-node-generations-testcase.xml:v-05',
        '3200-dimension-relationship-node/3250-dimension-relationship-node-generations/3250-dimension-relationship-node-generations-testcase.xml:v-05i',
        '3200-dimension-relationship-node/3250-dimension-relationship-node-generations/3250-dimension-relationship-node-generations-testcase.xml:v-06',
        '3200-dimension-relationship-node/3250-dimension-relationship-node-generations/3250-dimension-relationship-node-generations-testcase.xml:v-07',
        '3200-dimension-relationship-node/3250-dimension-relationship-node-generations/3250-dimension-relationship-node-generations-testcase.xml:v-08',
        '6000-aspect-node/6000-aspect-node-testcase.xml:v-01',
        '6000-aspect-node/6000-aspect-node-testcase.xml:v-02',
        '6000-aspect-node/6000-aspect-node-testcase.xml:v-03',
        '6000-aspect-node/6000-aspect-node-testcase.xml:v-04',
        '6000-aspect-node/6100-concept-aspect-node/6100-concept-aspect-node-testcase.xml:v-01',
        '6000-aspect-node/6100-concept-aspect-node/6100-concept-aspect-node-testcase.xml:v-01i',
        '6000-aspect-node/6200-period-aspect-node/6200-period-aspect-node-testcase.xml:v-01',
        '6000-aspect-node/6200-period-aspect-node/6200-period-aspect-node-testcase.xml:v-01i',
        '6000-aspect-node/6300-entity-identifier-aspect-node/6300-entity-identifier-aspect-node-testcase.xml:v-01',
        '6000-aspect-node/6300-entity-identifier-aspect-node/6300-entity-identifier-aspect-node-testcase.xml:v-01i',
        '6000-aspect-node/6400-unit-aspect-node/6400-unit-aspect-node-testcase.xml:v-01',
        '6000-aspect-node/6400-unit-aspect-node/6400-unit-aspect-node-testcase.xml:v-01i',
        '6000-aspect-node/6500-dimension-aspect-node/6500-dimension-aspect-node-testcase.xml:v-01',
        '6000-aspect-node/6500-dimension-aspect-node/6500-dimension-aspect-node-testcase.xml:v-01i',
        '6000-aspect-node/6500-dimension-aspect-node/6500-dimension-aspect-node-testcase.xml:v-02',
        '6000-aspect-node/6500-dimension-aspect-node/6500-dimension-aspect-node-testcase.xml:v-02i',
        '6000-aspect-node/6500-dimension-aspect-node/6500-dimension-aspect-node-testcase.xml:v-03',
        '6000-aspect-node/6500-dimension-aspect-node/6500-dimension-aspect-node-testcase.xml:v-03i',
        '6000-aspect-node/6500-dimension-aspect-node/6500-dimension-aspect-node-testcase.xml:v-05',
        '6000-aspect-node/6500-dimension-aspect-node/6500-dimension-aspect-node-testcase.xml:v-06',
        '6000-aspect-node/6600-aspect-node-concept-filter/6600-aspect-node-concept-filter-testcase.xml:v-01',
        '6000-aspect-node/6600-aspect-node-concept-filter/6600-aspect-node-concept-filter-testcase.xml:v-02',
        '6000-aspect-node/6610-aspect-node-period-filter/6610-aspect-node-period-filter-testcase.xml:v-01',
        '6000-aspect-node/6610-aspect-node-period-filter/6610-aspect-node-period-filter-testcase.xml:v-02',
        '6000-aspect-node/6610-aspect-node-period-filter/6610-aspect-node-period-filter-testcase.xml:v-03',
        '6000-aspect-node/6610-aspect-node-period-filter/6610-aspect-node-period-filter-testcase.xml:v-04',
        '6000-aspect-node/6610-aspect-node-period-filter/6610-aspect-node-period-filter-testcase.xml:v-05',
        '6000-aspect-node/6610-aspect-node-period-filter/6610-aspect-node-period-filter-testcase.xml:v-06',
        '6000-aspect-node/6620-aspect-node-entity-filter/6620-aspect-node-entity-filter-testcase.xml:v-01',
        '6000-aspect-node/6620-aspect-node-entity-filter/6620-aspect-node-entity-filter-testcase.xml:v-02',
        '6000-aspect-node/6630-aspect-node-unit-filter/6630-aspect-node-unit-filter-testcase.xml:v-01',
        '6000-aspect-node/6630-aspect-node-unit-filter/6630-aspect-node-unit-filter-testcase.xml:v-02',
        '6000-aspect-node/6630-aspect-node-unit-filter/6630-aspect-node-unit-filter-testcase.xml:v-03',
        '6000-aspect-node/6640-aspect-node-explicit-dimension-filter/6640-aspect-node-explicit-dimension-filter-testcase.xml:v-01',
        '6000-aspect-node/6640-aspect-node-explicit-dimension-filter/6640-aspect-node-explicit-dimension-filter-testcase.xml:v-02',
        '6000-aspect-node/6640-aspect-node-explicit-dimension-filter/6640-aspect-node-explicit-dimension-filter-testcase.xml:v-03',
        '6000-aspect-node/6650-aspect-node-typed-dimension-filter/6650-aspect-node-typed-dimension-filter-testcase.xml:v-01',
        '6000-aspect-node/6650-aspect-node-typed-dimension-filter/6650-aspect-node-typed-dimension-filter-testcase.xml:v-02',
        '6000-aspect-node/6650-aspect-node-typed-dimension-filter/6650-aspect-node-typed-dimension-filter-testcase.xml:v-03',
        '6000-aspect-node/6650-aspect-node-typed-dimension-filter/6650-aspect-node-typed-dimension-filter-testcase.xml:v-04',
        '6000-aspect-node/6650-aspect-node-typed-dimension-filter/6650-aspect-node-typed-dimension-filter-testcase.xml:v-04i',
        '6000-aspect-node/6660-aspect-node-aspect-cover-filter/6660-aspect-node-aspect-cover-filter-testcase.xml:v-01',
    ]),
    file='table-linkbase-conf-2015-08-12/conf/testcases-index.xml',
    info_url='https://specifications.xbrl.org/work-product-index-table-linkbase-table-linkbase-1.0.html',
    local_filepath='table-linkbase-conf-2015-08-12.zip',
    name=PurePath(__file__).stem,
    public_download_url='https://www.xbrl.org/2015/table-linkbase-conf-2015-08-12.zip'
)
